from pathlib import Path
import logging

import pytest
import numpy as np
import pandas as pd

import locan.data.metadata_pb2
from locan import ROOT_DIR, LocData
from locan.locan_io.locdata.io_locdata import load_txt_file, load_rapidSTORM_file
from locan.dependencies import HAS_DEPENDENCY, QtBindings, QT_BINDINGS


logger = logging.getLogger(__name__)

for key, value in HAS_DEPENDENCY.items():
    if not value:
        logger.info(f'Extra dependency {key} is not available.')

if QT_BINDINGS == QtBindings.NONE:
    logger.warning(f'Extra dependency for Qt bindings (pyside2 or pyqt5) is not available.')


# register pytest markers - should be in sync with pyproject.toml
def pytest_configure(config):
    config.addinivalue_line(
        "markers", "gui: marks tests that require user interaction (skipped by default)")
    config.addinivalue_line(
        "markers", "visual: marks tests that require visual inspection (skipped by default)")
    config.addinivalue_line(
        "markers", "requires_datasets: marks tests that require data in locan_datasets directory (skipped by default)")


# fixtures for random points

@pytest.fixture(scope='session')
def few_random_points():
    points = np.array([[0.066, 0.64], [0.92, 0.65], [0.11, 0.40], [0.20, 0.17], [0.75, 0.92],
                       [0.01, 0.12], [0.23, 0.54], [0.05, 0.25], [0.70, 0.73], [0.43, 0.16]])
    return points


# fixtures for LocData objects

@pytest.fixture(scope='session')
def locdata_empty():
    df = pd.DataFrame()
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope='session')
def locdata_single_localization():
    locdata_dict = {
        'position_x': np.array([1]),
        'position_y': np.array([1]),
        'frame': np.array([1]),
        'intensity': np.array([1]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope='session')
def locdata_single_localization_3d():
    locdata_dict = {
        'position_x': np.array([1]),
        'position_y': np.array([1]),
        'position_z': np.array([1]),
        'frame': np.array([1]),
        'intensity': np.array([1]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope='session')
def locdata_1d():
    locdata_dict = {
        'position_x': np.array([1, 1, 2, 3, 4, 5]),
        'frame': np.array([1, 2, 2, 4, 5, 6]),
        'intensity': np.array([100, 150, 110, 80, 105, 95]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope='session')
def locdata_2d():
    locdata_dict = {
        'position_x': np.array([1, 1, 2, 3, 4, 5]),
        'position_y': np.array([1, 5, 3, 6, 2, 5]),
        'frame': np.array([1, 2, 2, 4, 5, 6]),
        'intensity': np.array([100, 150, 110, 80, 105, 95]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope='session')
def locdata_2d_negative():
    locdata_dict = {
        'position_x': np.array([1, -1, 2, 3, 4, 5]),
        'position_y': np.array([1, 5, 3, 6, -2, 5]),
        'frame': np.array([1, 2, 2, 4, 5, 6]),
        'intensity': np.array([100, 150, 110, 80, 105, 95]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope='session')
def locdata_3d():
    locdata_dict = {
        'position_x': np.array([1, 1, 2, 3, 4, 5]),
        'position_y': np.array([1, 5, 3, 6, 2, 5]),
        'position_z': np.array([1, 2, 5, 4, 3, 2]),
        'frame': np.array([1, 2, 2, 4, 5, 6]),
        'intensity': np.array([100, 150, 110, 80, 105, 95]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope='session')
def locdata_non_standard_index():
    locdata_dict = {
        'position_x': np.array([1, 1, 2, 3, 4, 5]),
        'position_y': np.array([1, 5, 3, 6, 2, 5]),
        'frame': np.array([1, 2, 2, 4, 5, 6]),
        'intensity': np.array([100, 150, 110, 80, 105, 95]),
    }
    df = pd.DataFrame(locdata_dict)
    df.index = [2, 1, 5, 10, 100, 200]
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope='session')
def locdata_rapidSTORM_2d():
    path = Path(ROOT_DIR / 'tests/test_data/rapidSTORM_dstorm_data.txt')
    dat = load_rapidSTORM_file(path)
    dat.meta.creation_time.FromSeconds(1)
    return dat


@pytest.fixture(scope='session')
def locdata_blobs_2d():
    path = Path(ROOT_DIR / 'tests/test_data/five_blobs.txt')
    dat = load_txt_file(path)
    dat.meta.creation_time.FromSeconds(1)
    return dat


@pytest.fixture(scope='session')
def locdata_blobs_3d():
    path = Path(ROOT_DIR / 'tests/test_data/five_blobs_3D.txt')
    dat = load_txt_file(path)
    dat.meta.creation_time.FromSeconds(1)
    return dat


@pytest.fixture(scope='session')
def locdata_two_cluster_2d():
    points = np.array([[0.5, 0.5], [1, 0.6], [1.1, 1], [5, 5.6], [5.1, 6], [5.5, 5]])
    locdata_dict = {
        'position_x': points.T[0],
        'position_y': points.T[1],
        'cluster_label': np.array([1, 1, 1, 2, 2, 2])
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    locdata = LocData.from_dataframe(dataframe=df, meta=meta_)
    locdata.region = locdata.bounding_box.region
    return locdata


@pytest.fixture(scope='session')
def locdata_two_cluster_with_noise_2d():
    points = np.array([[0.5, 0.5], [1, 0.6], [1.1, 1], [5, 5.6], [5.1, 6], [5.5, 5], [100, 100]])
    locdata_dict = {
        'position_x': points.T[0],
        'position_y': points.T[1],
        'cluster_label': np.array([1, 1, 1, 2, 2, 2, -1])
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    locdata = LocData.from_dataframe(dataframe=df, meta=meta_)
    locdata.region = locdata.bounding_box.region
    return locdata



from io import StringIO


import locan.constants
from locan.locan_io import load_thunderstorm_file
from locan.locan_io.locdata.thunderstorm_io import load_thunderstorm_header


def test_get_correct_column_names_from_Thunderstorm_header():
    columns = load_thunderstorm_header(
        path=locan.ROOT_DIR / 'tests/test_data/Thunderstorm_dstorm_data.csv')
    assert (columns == ['original_index', 'frame', 'position_x', 'position_y', 'psf_sigma', 'intensity',
                        'local_background', 'local_background_sigma', 'chi_square', 'uncertainty'])

    file_like = StringIO("id,frame,x [nm],y [nm]\n"
                         "73897.0,2001.0,1320.109670647555,26344.7124618434")
    columns = load_thunderstorm_header(path=file_like)
    assert (columns == ['original_index', 'frame', 'position_x', 'position_y'])


def test_loading_Thunderstorm_file():
    dat = load_thunderstorm_file(path=locan.ROOT_DIR / 'tests/test_data/Thunderstorm_dstorm_data.csv',
                                    nrows=10)
    # print(dat.data.columns)
    assert (len(dat) == 10)

    file_like = StringIO("id,frame,x [nm],y [nm]\n"
                         "73897.0,2001.0,1320.109670647555,26344.7124618434")
    dat = load_thunderstorm_file(path=file_like, nrows=1)
    assert (len(dat) == 1)



"""
Analyze the distribution of a localization property in 2D.

Analyze the distribution of a localization property as function of two other localization properties in 2D.
E.g. looking at how the local background is distributed over localization coordinates helps to characterize the
illumination profile in SMLM experiments.
"""
from collections import namedtuple
import logging

import numpy as np
import matplotlib.pyplot as plt
from lmfit import Model, Parameters

from locan.render.render2d import render_2d_mpl
from locan.analysis.analysis_base import _Analysis, _list_parameters
from locan.render.render2d import histogram
from locan.configuration import COLORMAP_DIVERGING


__all__ = ['LocalizationProperty2d']

logger = logging.getLogger(__name__)


##### The algorithms

# model fit function in 2d
def _gauss_2d(x, y, amplitude, center_x, center_y, sigma_x, sigma_y):
    """
    2D Gauss function of variables (x, y).

    Parameters
    ----------
    x : float, array-like
        all x values
    y : float, array-like
        all y values
    amplitude : float
        amplitude
    center_x : float
        x-shift
    center_y : float
        y-shift
    sigma_x : float
        standard deviation x
    sigma_y : float
        standard deviation y

    Returns
    -------
    float, numpy.ndarray
    """
    x_ = np.asarray(x)
    y_ = np.asarray(y)
    amplitude, center_x, center_y, sigma_x, sigma_y = (np.float64(number) for number in
                                                       (amplitude, center_x, center_y, sigma_x, sigma_y))
    return amplitude * np.exp(-((x_-center_x)**2/(2.*sigma_x**2) + (y_-center_y)**2/(2.*sigma_y**2)))


def _fit_image(data, bin_range):
    """
    Fit 2D Gauss function to image data.

    Parameters
    ----------
    data : numpy.ndarray of shape (3, n_image_values)
        arrays with corresponding values for x, y, z
    bin_range : numpy.ndarray
        range as returned from histogram().

    Returns
    -------
    lmfit.model.ModelResult object
        The fit results.
    """
    # prepare 1D lmfit model from 2D model function
    def model_function(points, amplitude=1, center_x=0, center_y=0, sigma_x=1, sigma_y=1):
        return np.ravel(_gauss_2d(*points.T, amplitude, center_x, center_y, sigma_x, sigma_y))

    model = Model(model_function, nan_policy='omit')
    # print(model.param_names, model.independent_vars)

    # instantiate lmfit Parameters
    params = Parameters()
    params.add('amplitude', value=np.amax(data[2]))
    centers = np.add(bin_range[:, 0], np.diff(bin_range).flatten() / 2)
    params.add('center_x', value=centers[0])
    params.add('center_y', value=centers[1])
    sigmas = np.diff(bin_range).ravel() / 4
    params.add('sigma_x', value=sigmas[0])
    params.add('sigma_y', value=sigmas[1])

    # fit
    model_result = model.fit(data[2], points=data[0:2].T, params=params)

    return model_result


def _localization_property2d(locdata, loc_properties=None, other_property=None,
                             bins=None, n_bins=None, bin_size=10, bin_edges=None, bin_range=None,
                             rescale=None):
    # bin localization data
    img, bins_, label = histogram(locdata, loc_properties, other_property, bins, bin_size, bin_range, rescale)

    # prepare one-dimensional data
    xx, yy = np.meshgrid(bins_.bin_edges[0][1:], bins_.bin_edges[1][1:])

    # eliminate image zeros
    # positions = np.nonzero(img)
    positions = np.nonzero(~np.isnan(img))

    data_0 = xx[positions].flatten()
    data_1 = yy[positions].flatten()
    data_2 = img[positions].flatten()

    data = np.stack((data_0, data_1, data_2))

    model_result = _fit_image(data, np.array(bins_.bin_range))

    Results = namedtuple('Results', 'image bins label model_result')
    results = Results(img, bins_, label, model_result)
    return results


##### The specific analysis classes

class LocalizationProperty2d(_Analysis):
    """
    Analyze 2d distribution of histogram for two localization properties.

    Fit a two dimensional Gauss distribution.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    loc_properties : list, None
        Localization properties to be grouped into bins. If None The coordinate_values of locdata are used.
    other_property : str, None
        Localization property (columns in locdata.data) that is averaged in each pixel. If None localization counts are
        shown.
    bins : int, sequence, Bins, boost_histogram.axis.Axis, None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple, list, numpy.ndarray of float with shape (dimension, n_bin_edges), None
        Array of bin edges for all or each dimension.
    n_bins : int, list, tuple, numpy.ndarray, None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float, list, tuple, numpy.ndarray, None
        The size of bins in units of locdata coordinate units for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple, tuple of tuples of float with shape (dimension, 2), None, 'zero'
        The data bin_range to be taken into consideration for all or each dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    rescale : int, str, locan.constants.Trafo, callable, bool, None
        Transformation as defined in Trafo or by transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min intensities
        For 'equal' intensity values are rescaled by histogram equalization.

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : lmfit.model.ModelResult object
        Computed fit results.
    """
    def __init__(self, meta=None,
                 loc_properties=None, other_property='local_background',
                 bins=None, n_bins=None, bin_size=100, bin_edges=None, bin_range=None,
                 rescale=None,
                 ):
        super().__init__(meta=meta, loc_properties=loc_properties, other_property=other_property,
                         bins=bins, n_bins=n_bins, bin_size=bin_size, bin_edges=bin_edges, bin_range=bin_range,
                         rescale=rescale)
        self.results = None

    def compute(self, locdata=None):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data.

        Returns
        -------
        Analysis class
            Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        self.results = _localization_property2d(locdata=locdata, **self.parameter)
        return self

    def report(self):
        if not self:
            logger.warning('No results available')
            return

        print('Fit results for:\n')
        print(self.results.model_result.fit_report(min_correl=0.25))
        # print(self.results.fit_results.best_values)

        # judge fit parameter
        max_fit_value = self.results.model_result.best_fit.max()
        min_fit_value = self.results.model_result.best_fit.min()
        ratio = (max_fit_value - min_fit_value) / max_fit_value
        print(f'Maximum fit value in image: {max_fit_value:.3f}')
        print(f'Minimum fit value in image: {min_fit_value:.3f}')
        print(f'Fit value variation over image range: {ratio:.2f}')

    def plot(self, ax=None, **kwargs):
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing plot(results).

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.contour`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        ax.imshow(self.results.image, cmap='viridis', origin='lower', extent=np.ravel(self.results.bins.bin_range))

        x, y = self.results.bins.bin_edges[0][1:], self.results.bins.bin_edges[1][1:]
        xx, yy = np.meshgrid(x, y)
        zz = np.stack((xx, yy), axis=-1).reshape((np.product(xx.shape), 2))
        z = self.results.model_result.eval(points=zz)

        contourset = ax.contour(x, y, z.reshape((len(y), len(x))), 8, colors='w', **kwargs)
        plt.clabel(contourset, fontsize=9, inline=1)
        ax.set(title=self.parameter['other_property'],
               xlabel=self.results.label[0],
               ylabel=self.results.label[1]
               )

        return ax

    def plot_residuals(self, ax=None, **kwargs):
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing plot(results).

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.contour`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        x, y = self.results.bins.bin_edges[0][1:], self.results.bins.bin_edges[1][1:]
        xx, yy = np.meshgrid(x, y)
        zz = np.stack((xx, yy), axis=-1).reshape((np.product(xx.shape), 2))
        z = self.results.model_result.eval(points=zz)

        residuals = np.where(self.results.image == 0, np.nan, z.reshape((len(y), len(x))) - self.results.image)
        max_absolute_value = max([abs(np.nanmin(residuals)), abs(np.nanmax(residuals))])
        ax.imshow(residuals, cmap=COLORMAP_DIVERGING, origin='lower', extent=np.ravel(self.results.bins.bin_range),
                  vmin=(-max_absolute_value), vmax=max_absolute_value)

        # contourset = ax.contour(x, y, z.special((len(y), len(x))), 8, colors='w', **kwargs)
        # plt.clabel(contourset, fontsize=9, inline=1)
        ax.set(title=self.parameter['other_property'],
               xlabel=self.results.label[0],
               ylabel=self.results.label[1]
               )

        return ax

    def plot_deviation_from_mean(self, ax=None, **kwargs):
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing plot(results).

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.contour`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        positions = np.nonzero(self.results.image)
        mean_value = self.results.image[positions].mean()
        deviations = np.where(self.results.image == 0, np.nan, self.results.image - mean_value)
        max_absolute_value = max([abs(np.nanmin(deviations)), abs(np.nanmax(deviations))])
        ax.imshow(deviations, cmap=COLORMAP_DIVERGING, origin='lower', extent=np.ravel(self.results.bins.bin_range),
                  vmin=(-max_absolute_value), vmax=max_absolute_value)

        ax.set(title=f"{self.parameter['other_property']} - deviation from mean",
               xlabel=self.results.label[0],
               ylabel=self.results.label[1]
               )

        return ax

    def plot_deviation_from_median(self, ax=None, **kwargs):
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing plot(results).

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes on which to show the image
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.contour`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        positions = np.nonzero(self.results.image)
        median_value = np.median(self.results.image[positions])
        deviations = np.where(self.results.image == 0, np.nan, self.results.image - median_value)
        max_absolute_value = max([abs(np.nanmin(deviations)), abs(np.nanmax(deviations))])
        ax.imshow(deviations, cmap=COLORMAP_DIVERGING, origin='lower', extent=np.ravel(self.results.bins.bin_range),
                  vmin=(-max_absolute_value), vmax=max_absolute_value)

        ax.set(title=f"{self.parameter['other_property']} - deviation from median",
               xlabel=self.results.label[0],
               ylabel=self.results.label[1]
               )

        return ax



import pytest
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt  # needed for visual inspection

import locan.data.metadata_pb2
from locan import render_2d_mpl  # needed for visual inspection
from locan import LocData, scatter_2d_mpl, transform_affine, cluster_dbscan, HullType
from locan import Rectangle, RoiRegion
from locan import select_by_condition, random_subset, select_by_region, \
    exclude_sparse_points, localizations_in_cluster_regions


@pytest.fixture()
def locdata_simple_():
    locdata_dict = {
        'position_x': [0, 1, 2, 3, 0, 1, 4, 5],
        'position_y': [0, 1, 2, 3, 1, 4, 5, 1],
        'position_z': [0, 1, 2, 3, 4, 4, 4, 5]
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def locdata_simple():
    locdata_dict = {
        'position_x': [0, 1, 2, 3, 0, 1, 4, 5],
        'position_y': [0, 1, 2, 3, 1, 4, 5, 1],
        'position_z': [0, 1, 2, 3, 4, 4, 4, 5]
    }
    df = pd.DataFrame(locdata_dict)
    df.index = [2, 0, 1, 3, 4, 5, 6, 7]
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.FromSeconds(1)
    return LocData.from_dataframe(dataframe=df, meta=meta_)


def test_select_by_condition(locdata_simple):
    dat_s = select_by_condition(locdata_simple, 'position_x>1')
    assert (len(dat_s) == 4)
    assert np.all(dat_s.data.index == [1, 3, 6, 7])
    # dat_s.print_meta()
    # print(dat_s.meta)
    # print(dat_s.data)


def test_LocData_selection_from_collection(locdata_simple):
    # print(locdata_simple.meta)
    sel = []
    for i in range(4):
        sel.append(select_by_condition(locdata_simple, f'position_x>{i}'))
    col = LocData.from_collection(sel)
    assert (len(col) == 4)
    assert (len(col.references) == 4)
    # print(col.references[0].meta)
    # print(col.data)
    # print(col.meta)

    col_sel = select_by_condition(col, 'localization_count>2')
    assert (len(col_sel) == 3)
    # print(col_sel.data)
    assert(col_sel.references is col)

    col_sel_sel = select_by_condition(col_sel, 'localization_count<4')
    assert (len(col_sel_sel) == 1)
    # print(col_sel_sel.data)
    assert(col_sel_sel.references is col_sel)
    # print(col_sel_sel.meta)


def test_random_subset(locdata_simple):
    dat_s = random_subset(locdata_simple, n_points=3)
    assert (len(dat_s) == 3)
    # dat_s.print_meta()
    # print(dat_s.data)
    # print(dat_s.meta)


def test_localizations_in_cluster_regions(locdata_blobs_2d):
    coordinates = [(200, 500), (200, 600), (900, 650), (1000, 600)]
    locdata = LocData.from_coordinates(coordinates)
    noise, collection = cluster_dbscan(locdata_blobs_2d, eps=100, min_samples=3)
    # print(collection.data)

    # visualize
    # ax = render_2d_mpl(locdata_blobs_2d)
    # scatter_2d_mpl(collection.references[2], index=False, marker='.', color='g')
    # scatter_2d_mpl(locdata, index=False, marker='o')
    # scatter_2d_mpl(collection)
    # ax.add_patch(collection.references[2].convex_hull.region.as_artist(fill=False))
    # plt.show()

    # print(locdata_blobs_2d.convex_hull.region.contains(coordinates))
    # print(collection.references[2].convex_hull.region.contains(coordinates))

    # collection with references being a list of other LocData objects, e.g. individual clusters
    result = localizations_in_cluster_regions(locdata, collection)
    assert np.array_equal(result.data.localization_count.values, [0, 0, 1, 0, 0])

    result = localizations_in_cluster_regions(locdata, collection, hull_type=HullType.BOUNDING_BOX)
    assert np.array_equal(result.data.localization_count.values, [0, 0, 1, 0, 1])

    # selection of collection with references being another LocData object
    selected_collection = select_by_condition(collection, condition='subregion_measure_bb > 200')
    result = localizations_in_cluster_regions(locdata, selected_collection)
    assert np.array_equal(result.data.localization_count.values, [0, 1, 0, 0])

    # collection being a list of other LocData objects
    result = localizations_in_cluster_regions(locdata, collection.references)
    assert np.array_equal(result.data.localization_count.values, [0, 0, 1, 0, 0])

    plt.close('all')


def test_select_by_region(locdata_2d):
    region = Rectangle((1, 1), 3.5, 4.5, 0)

    # visualize
    # ax = scatter_2d_mpl(locdata_2d, index=True, marker='o', color='g')
    # ax.add_patch(region.as_artist(fill=False))
    # plt.show()

    new_locdata = select_by_region(locdata_2d, region)
    # print(new_locdata.data)
    assert new_locdata.meta.history[-1].name == "select_by_region"
    assert len(new_locdata) == 2

    new_locdata = select_by_region(locdata_2d, region, loc_properties=['position_x', 'frame'])
    assert len(new_locdata) == 3

    region = RoiRegion(region_type='rectangle', region_specs=((1, 1), 3.5, 4.5, 0))
    new_locdata = select_by_region(locdata_2d, region, loc_properties=['position_x', 'frame'])
    assert len(new_locdata) == 3

    plt.close('all')


def test_exclude_sparse_points(locdata_simple):
    new_locdata = exclude_sparse_points(locdata=locdata_simple, radius=3, min_samples=2)
    assert len(new_locdata) == 4
    # todo: check if the correct points are taken
    locdata_simple_trans = transform_affine(locdata_simple, offset=(0.5, 0, 0))
    new_locdata = exclude_sparse_points(locdata=locdata_simple,
                                        other_locdata=locdata_simple_trans, radius=2, min_samples=2)
    assert len(new_locdata) == 3
    # print(new_locdata.meta)



#!/usr/bin/env python

"""
Define regions of interest with matplotlib and save as roi files.

With this script you can choose a file name, open the localization file and draw a rectangular region of interest.
Within the matplotlib image draw a rectange, type '+' to add the roi to the list, then type 'q' to quit.
The roi is then saved as _roi.yaml file.

To run the script::

    locan draw_roi_mpl -f <file> -t <file type> -i <roi file indicator> -r <region type>

Try for instance::

    locan draw_roi_mpl -f "locan/tests/test_data/five_blobs.txt" -t 1 -i "_roi" -r "ellipse"

See Also
--------
locan.data.rois.select_by_drawing_mpl

"""

import argparse
from pathlib import Path

from locan.gui.io import file_dialog
import locan.locan_io.locdata.io_locdata as io
from locan.render.render2d import select_by_drawing_mpl


def sc_draw_roi_mpl(file_path=None, file_type=1, roi_file_indicator='_roi', region_type='rectangle'):
    """
    Define regions of interest by drawing a boundary.

    Parameters
    ----------
    file_path : str, bytes, os.PathLike
        File path to localization data.
    file_type : int, str, locan.constants.FileType, locan.data.metadata_pb2.Metadata
        Indicator for the file type.
        Integer or string should be according to locan.constants.FileType.
    roi_file_indicator : str
        Indicator to add to the localization file name and use as roi file name (with further extension .yaml).
    region_type : str
        rectangle, ellipse, or polygon specifying the selection widget to use.
    """

    # choose file interactively
    if file_path is None:
        file_path = Path(file_dialog(message='choose file', filter='*.txt; *.csv')[0])
    else:
        file_path = Path(file_path)

    print(file_path)

    # load data
    dat = io.load_locdata(path=file_path, file_type=file_type)

    # set roi
    rois = select_by_drawing_mpl(locdata=dat, bin_size=50, rescale='equal', region_type=region_type)
    print(rois)

    # save roi
    for i, roi in enumerate(rois):
        roi_file = file_path.stem + roi_file_indicator + f'_{i}.yaml'
        roi_path = file_path.with_name(roi_file)
        roi.to_yaml(path=roi_path)


def _add_arguments(parser):
    parser.add_argument('-f', '--file', dest='file', type=str, default=None,
                        help='File path to localization data.')
    parser.add_argument('-t', '--type', dest='type', type=int, default=1,
                        help='Integer or string indicating the file type.')
    parser.add_argument('-i', '--indicator', dest='roi_file_indicator', type=str, default='_roi',
                        help='Indicator to add to the localization file name and use as roi file name '
                             '(with further extension .yaml).')
    parser.add_argument('-r', '--region', dest='region', type=str, default='rectangle',
                        help='String indicating the region type.')


def main(args=None):

    parser = argparse.ArgumentParser(description='Set roi by drawing a boundary.')
    _add_arguments(parser)
    returned_args = parser.parse_args(args)

    sc_draw_roi_mpl(returned_args.file, returned_args.type, returned_args.roi_file_indicator,
                    returned_args.region_type)


if __name__ == '__main__':
    main()



import logging

import pytest
import matplotlib.pyplot as plt
import numpy as np
from shapely import affinity
from shapely.geometry import MultiPoint

from locan import scatter_2d_mpl
from locan.simulation import make_uniform, make_cluster
from locan import distance_to_region, distance_to_region_boundary, max_distance, inertia_moments
from locan import regions_union, RoiRegion, Rectangle, Ellipse


def test_distance_to_region_RoiRegion(locdata_2d):
    region = RoiRegion(region_type='rectangle', region_specs=((1, 1), 3, 3, 0))
    # visualize
    ax = scatter_2d_mpl(locdata_2d, index=True, marker='o', color='g')
    ax.add_patch(region.as_artist(fill=False))
    # plt.show()
    distances = distance_to_region(locdata_2d, region)
    assert np.array_equal(distances[:-1], np.array([0, 1, 0, 2, 0]))

    region_2 = RoiRegion(region_type='rectangle', region_specs=((3, 3), 3, 3, 0))
    distances = distance_to_region(locdata_2d, regions_union([region, region_2]))
    assert np.array_equal(distances[:-1], np.array([0, 1, 0, 0, 0]))

    plt.close('all')


def test_distance_to_region(locdata_2d):
    region = Rectangle((1, 1), 3, 3, 0)
    # visualize
    ax = scatter_2d_mpl(locdata_2d, index=True, marker='o', color='g')
    ax.add_patch(region.as_artist(fill=False))
    # plt.show()
    distances = distance_to_region(locdata_2d, region)
    assert np.array_equal(distances[:-1], np.array([0, 1, 0, 2, 0]))

    region_2 = Rectangle((3, 3), 3, 3, 0)
    distances = distance_to_region(locdata_2d, regions_union([region, region_2]))
    assert np.array_equal(distances[:-1], np.array([0, 1, 0, 0, 0]))

    plt.close('all')


def test_distance_to_region_boundary_RoiRegion(locdata_2d):
    region = RoiRegion(region_type='rectangle', region_specs=((1, 1), 3, 3, 0))
    # visualize
    # ax = scatter_2d_mpl(locdata_2d, index=True, marker='o', color='g')
    # ax.add_patch(region.as_artist(fill=False))
    # plt.show()
    distances = distance_to_region_boundary(locdata_2d, region)
    assert np.array_equal(distances[:-1], np.array([0, 1, 1, 2, 0]))

    region_2 = RoiRegion(region_type='rectangle', region_specs=((3.5, 3.5), 3, 3, 0))
    distances = distance_to_region(locdata_2d, regions_union([region, region_2]))
    assert np.array_equal(distances[:-1], np.array([0, 1, 0, 0.5, 0]))

    plt.close('all')


def test_distance_to_region_boundary(locdata_2d):
    region = Rectangle((1, 1), 3, 3, 0)
    # visualize
    # ax = scatter_2d_mpl(locdata_2d, index=True, marker='o', color='g')
    # ax.add_patch(region.as_artist(fill=False))
    # plt.show()
    distances = distance_to_region_boundary(locdata_2d, region)
    assert np.array_equal(distances[:-1], np.array([0, 1, 1, 2, 0]))

    region_2 = RoiRegion(region_type='rectangle', region_specs=((3.5, 3.5), 3, 3, 0))
    distances = distance_to_region(locdata_2d, regions_union([region, region_2]))
    assert np.array_equal(distances[:-1], np.array([0, 1, 0, 0.5, 0]))

    plt.close('all')


def test_max_distance_2d(locdata_2d):
    mdist = max_distance(locdata=locdata_2d)
    assert (mdist == {'max_distance': 5.656854249492381})


def test_max_distance_3d(locdata_3d):
    mdist = max_distance(locdata=locdata_3d)
    assert (mdist == {'max_distance': 6.164414002968976})


@pytest.mark.visual
def test_compute_inertia_moments_visual():
    rng = np.random.default_rng(seed=1)
    offspring = [make_uniform(n_samples=10, region=Ellipse((0, 0), 1, 1), seed=rng)] * 2
    points, _, _, _ = make_cluster(centers=([0, 0], [10, 0]), region=((0, 10), (0, 10)), offspring=offspring,
                                clip=False, seed=rng)

    fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(10, 5))
    for angle in np.linspace(0, 180, 20):
        rotated_points = affinity.rotate(MultiPoint(points), angle, origin=[0, 0], use_radians=False)
        rotated_points = np.array([rpts.coords[0] for rpts in rotated_points.geoms])
        axes[0].scatter(*rotated_points.T)
        inertia_moments_ = inertia_moments(rotated_points)
        axes[1].scatter(angle, inertia_moments_.orientation)
    plt.show()


def test_compute_inertia_moments():
    rng = np.random.default_rng(seed=1)
    offspring = [make_uniform(n_samples=100, region=Ellipse((0, 0), 1, 1), seed=rng)] * 2
    points, _, _, _ = make_cluster(centers=([0, 0], [10, 0]), region=((0, 10), (0, 10)), offspring=offspring,
                                clip=False, seed=rng)
    for angle in np.linspace(0, 180, 5):
        rotated_points = affinity.rotate(MultiPoint(points), angle, origin=[0, 0], use_radians=False)
        rotated_points = np.array([rpts.coords[0] for rpts in rotated_points.geoms])
        inertia_moments_ = inertia_moments(rotated_points)
        assert len(inertia_moments_.eigenvalues) == 2
        assert len(inertia_moments_.eigenvectors) == 2
        assert len(inertia_moments_.variance_explained) == 2

        assert (inertia_moments_.orientation + 180) % 180 == pytest.approx((angle + 180.000001) % 180, abs=0.01)
        assert (inertia_moments_.orientation+180) % 180 == pytest.approx((angle+180.000001) % 180, abs=0.01)
        assert inertia_moments_.eccentricity == pytest.approx(0.98, abs=0.1)


def test_compute_inertia_moments_3d(caplog):
    rng = np.random.default_rng(seed=1)
    offspring = [rng.normal(size=(100, 3), scale=1)] * 2
    points, _, _, _ = make_cluster(centers=([0, 0, 0], [10, 0, 0]), region=((0, 10), (0, 10), (0, 10)), offspring=offspring,
                                clip=False, seed=rng)

    inertia_moments_ = inertia_moments(np.array(points))
    assert len(inertia_moments_.eigenvalues) == 3
    assert np.isnan(inertia_moments_.orientation)
    assert np.isnan(inertia_moments_.eccentricity)
    assert caplog.record_tuples == [('locan.data.properties.misc', logging.WARNING,
                                     'Orientation and eccentricity have not yet been implemented for 3D.')]



"""

This module provides functions for rendering locdata objects in 2D.

"""
import logging

import numpy as np
from matplotlib import pyplot as plt
import matplotlib.colors as mcolors
from skimage import exposure
import scipy.signal.windows

from locan import locdata_id
from locan.data import LocData
from locan.data.rois import Roi
from locan.constants import RenderEngine
from locan.configuration import COLORMAP_CONTINUOUS, RENDER_ENGINE
from locan.dependencies import HAS_DEPENDENCY
from locan.data.rois import _MplSelector
from locan.data.aggregate import histogram, Bins, _check_loc_properties
from locan.data.properties.locdata_statistics import ranges
from locan.render.utilities import _napari_shape_to_region
from locan.render.transform import adjust_contrast

if HAS_DEPENDENCY["mpl_scatter_density"]: import mpl_scatter_density
if HAS_DEPENDENCY["napari"]: import napari


__all__ = ['render_2d', 'render_2d_mpl', 'render_2d_scatter_density', 'render_2d_napari', 'scatter_2d_mpl',
           'apply_window', 'select_by_drawing_napari', 'render_2d_rgb_mpl', 'render_2d_rgb_napari']

logger = logging.getLogger(__name__)


def render_2d_mpl(locdata, loc_properties=None, other_property=None,
                  bins=None, n_bins=None, bin_size=10, bin_edges=None, bin_range=None,
                  rescale=None,
                  ax=None, cmap=COLORMAP_CONTINUOUS, cbar=True, colorbar_kws=None,
                  interpolation='nearest', **kwargs):
    """
    Render localization data into a 2D image by binning x,y-coordinates into regular bins.

    Parameters
    ----------
    locdata : LocData
        Localization data.
    loc_properties : list, None
        Localization properties to be grouped into bins. If None The coordinate_values of locdata are used.
    other_property : str, None
        Localization property (columns in locdata.data) that is averaged in each pixel. If None localization counts are
        shown.
    bins : int, sequence, Bins, boost_histogram.axis.Axis, None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple, list, numpy.ndarray of float with shape (dimension, n_bin_edges), None
        Array of bin edges for all or each dimension.
    n_bins : int, list, tuple, numpy.ndarray, None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float, list, tuple, numpy.ndarray, None
        The size of bins in units of locdata coordinate units for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple, tuple of tuples of float with shape (dimension, 2), None, 'zero'
        The data bin_range to be taken into consideration for all or each dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    rescale : int, str, locan.Trafo, callable, bool, None
        Transformation as defined in Trafo or by transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min intensities
        For 'equal' intensity values are rescaled by histogram equalization.
    ax : matplotlib.axes.Axes
        The axes on which to show the image
    cmap : str or Colormap instance
        The colormap used to map normalized data values to RGBA colors.
    cbar : bool
        If true draw a colorbar. The colobar axes is accessible using the cax property.
    colorbar_kws : dict
        Keyword arguments for :func:`matplotlib.pyplot.colorbar`.
    interpolation : str
        Keyword argument for :func:`matplotlib.axes.Axes.imshow`.
    kwargs : dict
        Other parameters passed to :func:`matplotlib.axes.Axes.imshow`.

    Returns
    -------
    matplotlib.axes.Axes
        Axes object with the image.
    """
    # Provide matplotlib.axes.Axes if not provided
    if ax is None:
        ax = plt.gca()

    # return ax if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning('Locdata carries a single localization.')
        return ax

    data, bins, labels = histogram(locdata, loc_properties, other_property,
                                   bins, n_bins, bin_size, bin_edges, bin_range
                                   )
    data = adjust_contrast(data, rescale)

    mappable = ax.imshow(data.T, origin='lower', extent=[*bins.bin_range[0], *bins.bin_range[1]],
                         cmap=cmap, interpolation=interpolation, **kwargs)

    ax.set(
        title=labels[-1],
        xlabel=labels[0],
        ylabel=labels[1]
        )

    if cbar:
        if colorbar_kws is None:
            plt.colorbar(mappable, ax=ax)
        else:
            plt.colorbar(mappable, **colorbar_kws)

    return ax


def render_2d_scatter_density(locdata, loc_properties=None, other_property=None, bin_range=None,
                              ax=None, cmap=COLORMAP_CONTINUOUS, cbar=True, colorbar_kws=None, **kwargs):
    """
    Render localization data into a 2D image by binning x,y-coordinates into regular bins.

    Prepare :class:`matplotlib.axes.Axes` with image.

    Note
    ----
    To rescale intensity values use norm keyword.

    Parameters
    ----------
    locdata : LocData
        Localization data.
    loc_properties : list or None
        Localization properties to be grouped into bins. If None The coordinate_values of locdata are used.
    other_property : str or None
        Localization property (columns in locdata.data) that is averaged in each pixel. If None localization counts are
        shown.
    bin_range : tuple with shape (dimension, 2) or None or 'zero'
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    ax : matplotlib.axes.Axes
        The axes on which to show the image
    cmap : str or Colormap instance
        The colormap used to map normalized data values to RGBA colors.
    cbar : bool
        If true draw a colorbar. The colobar axes is accessible using the cax property.
    colorbar_kws : dict
        Keyword arguments for :func:`matplotlib.pyplot.colorbar`.
    kwargs : dict
        Other parameters passed to :class:`mpl_scatter_density.ScatterDensityArtist`.

    Returns
    -------
    matplotlib.axes.Axes
        Axes object with the image.
    """
    if not HAS_DEPENDENCY["mpl_scatter_density"]:
        raise ImportError('mpl-scatter-density is required.')

    # Provide matplotlib.axes.Axes if not provided
    if ax is None:
        ax = plt.gca()

    # return ax if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning('Locdata carries a single localization.')
        return ax
    else:
        fig = ax.get_figure()
        ax = fig.add_subplot(1, 1, 1, projection='scatter_density', label='scatter_density')

    if loc_properties is None:
        data = locdata.coordinates.T
        labels = list(locdata.coordinate_labels)
    elif isinstance(loc_properties, str) and loc_properties in locdata.coordinate_labels:
        data = locdata.data[loc_properties].values.T
        labels = list(loc_properties)
    elif isinstance(loc_properties, (list, tuple)):
        for prop in loc_properties:
            if prop not in locdata.coordinate_labels:
                raise ValueError(f'{prop} is not a valid property in locdata.')
        data = locdata.data[list(loc_properties)].values.T
        labels = list(loc_properties)
    else:
        raise ValueError(f'{loc_properties} is not a valid property in locdata.')

    if bin_range is None or isinstance(bin_range, str):
        bin_range_ = ranges(locdata, loc_properties=labels, special=bin_range)
    else:
        bin_range_ = bin_range

    if other_property is None:
        # histogram data by counting points
        if data.shape[0] == 2:
            values = None
        else:
            raise TypeError('Only 2D data is supported.')
        labels.append('counts')
    elif other_property in locdata.data.columns:
        # histogram data by averaging values
        if data.shape[0] == 2:
            # here color serves as weight since it is averaged over all points before binning.
            values = locdata.data[other_property].values.T
        else:
            raise TypeError('Only 2D data is supported.')
        labels.append(other_property)
    else:
        raise TypeError(f'No valid property name {other_property}.')

    a = mpl_scatter_density.ScatterDensityArtist(ax, *data, c=values, origin='lower', extent=[*bin_range_[0], *bin_range_[1]],
                                                 cmap=cmap, **kwargs)
    mappable = ax.add_artist(a)
    ax.set_xlim(*bin_range_[0])
    ax.set_ylim(*bin_range_[1])

    ax.set(title=labels[-1],
           xlabel=labels[0],
           ylabel=labels[1]
           )

    if cbar:
        if colorbar_kws is None:
            plt.colorbar(mappable, ax=ax, label=labels[-1])
        else:
            plt.colorbar(mappable, **colorbar_kws)

    return ax


def render_2d_napari(locdata, loc_properties=None, other_property=None,
                     bins=None, n_bins=None, bin_size=10, bin_edges=None, bin_range=None,
                     rescale=None, viewer=None, cmap='viridis', **kwargs):
    """
    Render localization data into a 2D image by binning x,y-coordinates into regular bins.
    Render the data using napari.

    Parameters
    ----------
    locdata : LocData
        Localization data.
    loc_properties : list, None
        Localization properties to be grouped into bins. If None The coordinate_values of `locdata` are used.
    other_property : str, None
        Localization property (columns in `locdata.data`) that is averaged in each pixel.
        If None, localization counts are shown.
    bins : int, sequence, Bins, boost_histogram.axis.Axis, None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple, list, numpy.ndarray[float] with shape (dimension, n_bin_edges), None
        Array of bin edges for all or each dimension.
    n_bins : int, list, tuple, numpy.ndarray, None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float, list, tuple, numpy.ndarray, None
        The size of bins in units of `locdata` coordinate units for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple, tuple[tuples[float]] with shape (dimension, 2), str, None
        The data bin_range to be taken into consideration for all or each dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    rescale : int, str, locan.Trafo, callable, bool, None
        Transformation as defined in :class:`locan.Trafo` or by transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min intensities
        For 'equal' intensity values are rescaled by histogram equalization.
    viewer : napari.Viewer
        The viewer object on which to add the image
    cmap : str, Colormap
        The Colormap object used to map normalized data values to RGBA colors.
    kwargs : dict
        Other parameters passed to :func:`napari.Viewer.add_image`.

    Returns
    -------
    napari.Viewer, Bins
    """
    if not HAS_DEPENDENCY["napari"]:
        raise ImportError('Function requires napari.')

    # Provide napari viewer if not provided
    if viewer is None:
        viewer = napari.Viewer()

    # return ax if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning('Locdata carries a single localization.')
        return viewer

    data, bins, labels = histogram(locdata, loc_properties, other_property,
                                   bins, n_bins, bin_size, bin_edges, bin_range
                                   )
    data = adjust_contrast(data, rescale)

    viewer.add_image(data, name=f'LocData {locdata_id}', colormap=cmap, **kwargs)
    return viewer, bins


def render_2d(locdata, render_engine=RENDER_ENGINE, **kwargs):
    """
    Wrapper function to render localization data into a 2D image.
    For complete signatures see render_2d_mpl or corresponding functions.
    """
    if render_engine == RenderEngine.MPL:
        return render_2d_mpl(locdata, **kwargs)
    elif HAS_DEPENDENCY["mpl_scatter_density"] and render_engine == RenderEngine.MPL_SCATTER_DENSITY:
        return render_2d_scatter_density(locdata, **kwargs)
    elif HAS_DEPENDENCY["napari"] and render_engine == RenderEngine.NAPARI:
        return render_2d_napari(locdata, **kwargs)
    else:
        raise NotImplementedError(f"render_2d is not implemented for {render_engine}.")


def scatter_2d_mpl(locdata, ax=None, index=True, text_kwargs=None, **kwargs):
    """
    Scatter plot of locdata elements with text marker for each element.

    Parameters
    ----------
    locdata : LocData
       Localization data.
    ax : matplotlib.axes.Axes
       The axes on which to show the plot
    index : bool
       Flag indicating if element indices are shown.
    text_kwargs : dict
       Keyword arguments for :func:`matplotlib.axes.Axes.text`.
    kwargs : dict
       Other parameters passed to :func:`matplotlib.axes.Axes.scatter`.

    Returns
    -------
    matplotlib.axes.Axes
       Axes object with the image.
    """
    if text_kwargs is None:
        text_kwargs = {}

    # Provide matplotlib.axes.Axes if not provided
    if ax is None:
        ax = plt.gca()

    # return ax if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning('Locdata carries a single localization.')
        return ax

    coordinates = locdata.coordinates
    ax.scatter(*coordinates.T, **dict({'marker': '+', 'color': 'grey'}, **kwargs))

    # plot element number
    if index:
        for centroid, marker in zip(coordinates, locdata.data.index.values):
            ax.text(*centroid, marker, **dict({'color': 'grey', 'size': 20}, **text_kwargs))

    ax.set(
           xlabel='position_x',
           ylabel='position_y'
           )

    return ax


def apply_window(image, window_function='tukey', **kwargs):
    """
    Apply window function to image.

    Parameters
    ----------
    image : numpy.ndarray
        Image
    window_function : str
        Window function to apply. One of 'tukey', 'hann' or any other in `scipy.signal.windows`.
    kwargs : dict
        Other parameters passed to the `scipy.signal.windows` window function.
    """
    window_func = getattr(scipy.signal.windows, window_function)
    windows = [window_func(M, **kwargs) for M in image.shape]

    result = image.astype('float64')
    result *= windows[0]
    result *= windows[1][:, None]

    return result


def select_by_drawing_mpl(locdata, region_type='rectangle', **kwargs):
    """
    Select region of interest from rendered image by drawing rois.

    Parameters
    ----------
    locdata : LocData
        The localization data from which to select localization data.
    region_type : str
        rectangle, or ellipse specifying the selection widget to use.
    kwargs :
        Other parameters as specified for :func:`render_2d`.

    Returns
    -------
    list[Roi]

    See Also
    --------
    :func:`locan.scripts.sc_draw_roi_mpl` : script for drawing rois
    matplotlib.widgets : selector functions
    """

    fig, ax = plt.subplots(nrows=1, ncols=1)
    render_2d_mpl(locdata, ax=ax, **kwargs)
    selector = _MplSelector(ax, type=region_type)
    plt.show()
    roi_list = [Roi(reference=locdata, region_specs=roi['region_specs'],
                    region=roi['region']) for roi in selector.rois]
    return roi_list


def select_by_drawing_napari(locdata, napari_run=True, **kwargs):
    """
    Select region of interest from rendered image by drawing rois in napari.

    Rois will be created from shapes in napari.viewer.layers['Shapes'].

    Parameters
    ----------
    locdata : LocData
        The localization data from which to select localization data.
    napari_run : bool
        If `True` napari.run is called (set to `False` for testing).
    kwargs : dict
        Other parameters passed to :func:`render_2d_napari`.

    Returns
    -------
    list[Roi]

    See Also
    --------
    :func:`locan.scripts.rois` : script for drawing rois
    """
    # select roi
    viewer, bins = render_2d_napari(locdata, **kwargs)
    if napari_run:
        napari.run()

    vertices = viewer.layers['Shapes'].data
    types = viewer.layers['Shapes'].shape_type

    regions = []
    for verts, typ in zip(vertices, types):
        regions.append(_napari_shape_to_region(verts, bins.bin_edges, typ))

    roi_list = [Roi(reference=locdata, region=reg) for reg in regions]
    return roi_list


def render_2d_rgb_mpl(locdatas, loc_properties=None, other_property=None,
                      bins=None, n_bins=None, bin_size=10, bin_edges=None, bin_range=None,
                      rescale=None,
                      ax=None,
                      interpolation='nearest', **kwargs):
    """
    Render localization data into a 2D RGB image by binning x,y-coordinates into regular bins.

    Note
    ----
    For rescale=False no normalization is carried out image intensities are clipped to (0, 1) for float value
    or (0, 255) for integer values according to the matplotlib.imshow behavior.
    For rescale=None we apply a normalization to (min, max) of all intensity values.
    For all other rescale options the normalization is applied to each individual image.

    Parameters
    ----------
    locdatas : list[LocData]
        Localization data.
    loc_properties : list, None
        Localization properties to be grouped into bins. If None The coordinate_values of `locdata` are used.
    other_property : str, None
        Localization property (columns in `locdata.data`) that is averaged in each pixel.
        If None, localization counts are shown.
    bins : int, sequence, Bins, boost_histogram.axis.Axis, None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple, list, numpy.ndarray[float] with shape (dimension, n_bin_edges), None
        Array of bin edges for all or each dimension.
    n_bins : int, list, tuple, numpy.ndarray, None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float, list, tuple, numpy.ndarray, None
        The size of bins in units of `locdata` coordinate units for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple, tuple[tuples[float]] with shape (dimension, 2), str, None
        The data bin_range to be taken into consideration for all or each dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    rescale : int, str, locan.Trafo, callable, bool, None
        Transformation as defined in :class:`locan.Trafo` or by transformation function.
        For False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min intensities
        For 'equal' intensity values are rescaled by histogram equalization.
    ax : matplotlib.axes.Axes
        The axes on which to show the image
    cmap : str or Colormap instance
        The colormap used to map normalized data values to RGBA colors.
    cbar : bool
        If true draw a colorbar. The colobar axes is accessible using the cax property.
    colorbar_kws : dict
        Keyword arguments for :func:`matplotlib.pyplot.colorbar`.
    interpolation : str
        Keyword argument for :func:`matplotlib.axes.Axes.imshow`.
    kwargs : dict
        Other parameters passed to :func:`matplotlib.axes.Axes.imshow`.

    Returns
    -------
    matplotlib.axes.Axes
        Axes object with the image.
    """
    # Provide matplotlib.axes.Axes if not provided
    if ax is None:
        ax = plt.gca()

    locdata_temp = LocData.concat(locdatas)

    # return ax if no or single point in locdata
    if len(locdata_temp) < 2:
        if len(locdata_temp) == 1:
            logger.warning('Locdata carries a single localization.')
        return ax

    if bin_edges is None:
        _, bins, labels = histogram(locdata_temp, loc_properties, other_property,
                                    bins, n_bins, bin_size, bin_edges, bin_range
                                    )
    else:
        labels = _check_loc_properties(locdata_temp, loc_properties)
        bins = Bins(bin_edges=bin_edges, labels=labels)

    imgs = [histogram(locdata, loc_properties, other_property, bin_edges=bins.bin_edges
                      ).data
            for locdata in locdatas
            ]

    if rescale is None:
        norm = mcolors.Normalize(vmin=np.min(imgs), vmax=np.max(imgs))
    else:
        norm = rescale
    imgs = [adjust_contrast(img, rescale=norm) for img in imgs]

    new = np.zeros_like(imgs[0])
    rgb_stack = np.stack([new] * 3, axis=2)

    for i, img in enumerate(imgs):
        rgb_stack[:, :, i] = img

    rgb_stack = np.transpose(rgb_stack, axes=(1, 0, 2))
    ax.imshow(rgb_stack, origin='lower', extent=[*bins.bin_range[0], *bins.bin_range[1]],
              interpolation=interpolation, **kwargs)

    ax.set(
        title=labels[-1],
        xlabel=labels[0],
        ylabel=labels[1]
    )

    return ax


def render_2d_rgb_napari(locdatas, loc_properties=None, other_property=None,
                      bins=None, n_bins=None, bin_size=10, bin_edges=None, bin_range=None,
                      rescale=None,
                      viewer=None,
                      **kwargs):
    """
    Render localization data into a 2D RGB image by binning x,y-coordinates into regular bins.

    Note
    ----
    For rescale=False no normalization is carried out image intensities are clipped to (0, 1) for float value
    or (0, 255) for integer values according to the matplotlib.imshow behavior.
    For rescale=None we apply a normalization to (min, max) of all intensity values.
    For all other rescale options the normalization is applied to each individual image.

    Parameters
    ----------
    locdatas : list[LocData]
        Localization data.
    loc_properties : list, None
        Localization properties to be grouped into bins. If None The coordinate_values of `locdata` are used.
    other_property : str, None
        Localization property (columns in `locdata.data`) that is averaged in each pixel.
        If None, localization counts are shown.
    bins : int, sequence, Bins, boost_histogram.axis.Axis, None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple, list, numpy.ndarray[float] with shape (dimension, n_bin_edges), None
        Array of bin edges for all or each dimension.
    n_bins : int, list, tuple, numpy.ndarray, None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float, list, tuple, numpy.ndarray, None
        The size of bins in units of `locdata` coordinate units for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple, tuple[tuples[float]] with shape (dimension, 2), str, None
        The data bin_range to be taken into consideration for all or each dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    rescale : int, str, locan.Trafo, callable, bool, None
        Transformation as defined in :class:`locan.Trafo` or by transformation function.
        For False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min intensities
        For 'equal' intensity values are rescaled by histogram equalization.
    viewer : napari.Viewer
        The viewer object on which to add the image
    cmap : str, Colormap
        The Colormap object used to map normalized data values to RGBA colors.
    kwargs : dict
        Other parameters passed to :func:`napari.Viewer.add_image`.

    Returns
    -------
    napari.Viewer, Bins
    """
    if not HAS_DEPENDENCY["napari"]:
        raise ImportError('Function requires napari.')

    # Provide napari viewer if not provided
    if viewer is None:
        viewer = napari.Viewer()

    locdata_temp = LocData.concat(locdatas)

    # return viewer if no or single point in locdata
    if len(locdata_temp) < 2:
        if len(locdata_temp) == 1:
            logger.warning('Locdata carries a single localization.')
        return viewer

    if bin_edges is None:
        _, bins, labels = histogram(locdata_temp, loc_properties, other_property,
                                    bins, n_bins, bin_size, bin_edges, bin_range
                                    )
    else:
        labels = _check_loc_properties(locdata_temp, loc_properties)
        bins = Bins(bin_edges=bin_edges, labels=labels)

    imgs = [histogram(locdata, loc_properties, other_property, bin_edges=bins.bin_edges
                      ).data
            for locdata in locdatas
            ]

    if rescale is None:
        norm = mcolors.Normalize(vmin=np.min(imgs), vmax=np.max(imgs))
    else:
        norm = rescale
    imgs = [adjust_contrast(img, rescale=norm) for img in imgs]

    new = np.zeros_like(imgs[0])
    rgb_stack = np.stack([new] * 3, axis=2)

    for i, img in enumerate(imgs):
        rgb_stack[:, :, i] = img

    rgb_stack = np.transpose(rgb_stack, axes=(1, 0, 2))
    viewer.add_image(rgb_stack, name=f'LocData {locdata_id}', rgb=True, **kwargs)
    return viewer



import math

import pytest
import numpy as np
import matplotlib.pyplot as plt  # needed for visualization
import matplotlib.tri as mtri  # needed for visualization
import networkx as nx

from locan import AlphaShape, AlphaComplex, RoiRegion, EmptyRegion, Region2D


@pytest.mark.visual
def test_AlphaComplex_visual(locdata_2d):
    points = locdata_2d.coordinates
    alpha_complex = AlphaComplex(points)

    alpha = 1.6
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type='all')
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(alpha, type='exterior')
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(alpha, type='interior')
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type='regular')
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(alpha, type='singular')

    lengths = [len(sim) for sim in [
        ac_simplices_all, ac_simplices_exterior, ac_simplices_interior, ac_simplices_regular, ac_simplices_singular]]
    assert lengths == [8, 2, 0, 6, 2]

    locdata_2d.data.plot(*locdata_2d.coordinate_labels, kind='scatter')
    for simp in ac_simplices_all:
        plt.plot(points[simp, 0], points[simp, 1], '-b')
    for simp in ac_simplices_interior:
        plt.plot(points[simp, 0], points[simp, 1], '--g')
    for simp in ac_simplices_regular:
        plt.plot(points[simp, 0], points[simp, 1], '--r')
    for simp in ac_simplices_singular:
        plt.plot(points[simp, 0], points[simp, 1], '--y')
    plt.show()


@pytest.mark.visual
def test_AlphaComplex_1_visual():
    points = np.array([[10, 10], [10, 14], [11, 11], [12, 11], [13, 10], [13, 13], [15, 14], [17, 16],
                       [10, 24]])

    alpha_complex = AlphaComplex(points)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type='all')
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(alpha, type='exterior')
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(alpha, type='interior')
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type='regular')
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(alpha, type='singular')

    H = alpha_complex.graph_from_lines(alpha, type='regular')

    fig, ax = plt.subplots(1, 3)
    nx.draw_networkx(H, points, with_labels=True, ax=ax[0])
    # plt.show()

    plt.scatter(*points.T)
    for simp in ac_simplices_all:
        ax[1].plot(points[simp, 0], points[simp, 1], '-b')
    for simp in ac_simplices_interior:
        ax[1].plot(points[simp, 0], points[simp, 1], '--g')
    for simp in ac_simplices_regular:
        ax[1].plot(points[simp, 0], points[simp, 1], '--r')
    for simp in ac_simplices_singular:
        ax[1].plot(points[simp, 0], points[simp, 1], '--y')
    # plt.show()

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='all')]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='exterior')]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='interior')]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='regular')]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='singular')]

    tri = mtri.Triangulation(*points.T, triangles=triangles_all)
    ax[2].triplot(tri, '-b')
    tri = mtri.Triangulation(*points.T, triangles=triangles_interior)
    ax[2].triplot(tri, '--g')
    tri = mtri.Triangulation(*points.T, triangles=triangles_regular)
    ax[2].triplot(tri, '--r')
    tri = mtri.Triangulation(*points.T, triangles=triangles_singular)
    ax[2].triplot(tri, '--y')
    plt.show()


@pytest.mark.visual
def test_AlphaComplex_2_visual():
    points = np.array([[10, 10], [10, 14], [11, 11], [12, 11], [13, 10], [13, 13], [15, 14], [17, 16],
                       [20, 20], [20, 24], [21, 21], [22, 21], [23, 20], [23, 23], [25, 24],
                       [10, 24]])

    alpha_complex = AlphaComplex(points)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type='all')
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(alpha, type='exterior')
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(alpha, type='interior')
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type='regular')
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(alpha, type='singular')

    H = alpha_complex.graph_from_lines(alpha, type='regular')

    fig, ax = plt.subplots(1, 3)
    nx.draw_networkx(H, points, with_labels=True, ax=ax[0])
    # plt.show()

    plt.scatter(*points.T)
    for simp in ac_simplices_all:
        ax[1].plot(points[simp, 0], points[simp, 1], '-b')
    for simp in ac_simplices_interior:
        ax[1].plot(points[simp, 0], points[simp, 1], '--g')
    for simp in ac_simplices_regular:
        ax[1].plot(points[simp, 0], points[simp, 1], '--r')
    for simp in ac_simplices_singular:
        ax[1].plot(points[simp, 0], points[simp, 1], '--y')
    # plt.show()

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='all')]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='exterior')]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='interior')]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='regular')]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='singular')]

    tri = mtri.Triangulation(*points.T, triangles=triangles_all)
    ax[2].triplot(tri, '-b')
    tri = mtri.Triangulation(*points.T, triangles=triangles_interior)
    ax[2].triplot(tri, '--g')
    tri = mtri.Triangulation(*points.T, triangles=triangles_regular)
    ax[2].triplot(tri, '--r')
    tri = mtri.Triangulation(*points.T, triangles=triangles_singular)
    ax[2].triplot(tri, '--y')
    plt.show()


@pytest.mark.visual
def test_AlphaComplex_3_visual():
    points = np.array([[10, 10], [10, 20], [20, 20], [20, 10], [10, 15], [15, 20], [20, 15], [15, 10],
                       [10, 12], [10, 18], [20, 18], [20, 12], [17, 20], [12, 10], [17, 10], [12, 20],
                       [11, 11], [11, 19], [19, 19], [19, 11], [11, 15], [15, 19], [19, 15], [15, 11],
                       [13, 11], [11.5, 12], [12, 18], [13, 19], [17, 19], [18, 18], [17, 11], [18, 12],
                       [15, 15]
                       ])

    alpha_complex = AlphaComplex(points)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type='all')
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(alpha, type='exterior')
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(alpha, type='interior')
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type='regular')
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(alpha, type='singular')

    H = alpha_complex.graph_from_lines(alpha, type='regular')

    fig, ax = plt.subplots(1, 3)
    nx.draw_networkx(H, points, with_labels=True, ax=ax[0])
    # plt.show()

    plt.scatter(*points.T)
    for simp in ac_simplices_all:
        ax[1].plot(points[simp, 0], points[simp, 1], '-b')
    for simp in ac_simplices_interior:
        ax[1].plot(points[simp, 0], points[simp, 1], '--g')
    for simp in ac_simplices_regular:
        ax[1].plot(points[simp, 0], points[simp, 1], '--r')
    for simp in ac_simplices_singular:
        ax[1].plot(points[simp, 0], points[simp, 1], '--y')
    # plt.show()

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='all')]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='exterior')]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='interior')]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='regular')]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='singular')]

    tri = mtri.Triangulation(*points.T, triangles=triangles_all)
    ax[2].triplot(tri, '-b')
    tri = mtri.Triangulation(*points.T, triangles=triangles_interior)
    ax[2].triplot(tri, '--g')
    tri = mtri.Triangulation(*points.T, triangles=triangles_regular)
    ax[2].triplot(tri, '--r')
    tri = mtri.Triangulation(*points.T, triangles=triangles_singular)
    ax[2].triplot(tri, '--y')
    plt.show()


def test_AlphaComplex_0():
    points = np.array([])
    assert np.size(points) == 0
    alpha_complex = AlphaComplex(points)
    assert len(alpha_complex.triangles) == 0
    assert len(alpha_complex.lines) == 0
    assert alpha_complex.dimension is None
    assert alpha_complex.get_alpha_complex_lines(alpha=alpha_complex.optimal_alpha(), type='exterior') == []
    assert alpha_complex.optimal_alpha() is None
    assert alpha_complex.alphas().size == 0
    # assert isinstance(alpha_complex.alpha_shape(alpha=2.2), AlphaShape)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type='all')
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(alpha, type='exterior')
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(alpha, type='interior')
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type='regular')
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(alpha, type='singular')

    lengths = [len(sim) for sim in [
        ac_simplices_all, ac_simplices_exterior, ac_simplices_interior, ac_simplices_regular, ac_simplices_singular]]
    assert lengths == [0, 0, 0, 0, 0]

    triangles_all = alpha_complex.get_alpha_complex_triangles(alpha, type='all')
    assert len(triangles_all) == 0

    H = alpha_complex.graph_from_lines(alpha, type='regular')
    assert isinstance(H, nx.Graph)
    H = alpha_complex.graph_from_triangles(alpha, type='regular')
    assert isinstance(H, nx.Graph)


def test_AlphaComplex_1():
    points = np.array([[10, 10], [10, 14], [11, 11], [12, 11], [13, 10], [13, 13], [15, 14], [17, 16],
                       [10, 24]])
    assert len(points) == 9
    alpha_complex = AlphaComplex(points)
    assert len(alpha_complex.triangles) == 11
    assert len(alpha_complex.lines) == 19
    assert alpha_complex.dimension == 2
    assert alpha_complex.get_alpha_complex_lines(alpha=alpha_complex.optimal_alpha(), type='exterior') == []
    assert alpha_complex.optimal_alpha() in alpha_complex.alphas()
    assert np.max(alpha_complex.alphas()) == np.inf
    assert isinstance(alpha_complex.alpha_shape(alpha=2.2), AlphaShape)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type='all')
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(alpha, type='exterior')
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(alpha, type='interior')
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type='regular')
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(alpha, type='singular')

    lengths = [len(sim) for sim in [
        ac_simplices_all, ac_simplices_exterior, ac_simplices_interior, ac_simplices_regular, ac_simplices_singular]]
    assert lengths == [13, 6, 5, 5, 3]

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='all')]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='exterior')]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='interior')]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='regular')]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='singular')]

    lengths = [len(sim) for sim in [
        triangles_all, triangles_exterior, triangles_interior, triangles_regular, triangles_singular]]
    assert lengths == [6, 5, 1, 4, 1]

    H = alpha_complex.graph_from_lines(alpha, type='regular')
    assert isinstance(H, nx.Graph)
    H = alpha_complex.graph_from_triangles(alpha, type='regular')
    assert isinstance(H, nx.Graph)

    alpha = 1
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type='all')
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(alpha, type='exterior')
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(alpha, type='interior')
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type='regular')
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(alpha, type='singular')

    lengths = [len(sim) for sim in [
        ac_simplices_all, ac_simplices_exterior, ac_simplices_interior, ac_simplices_regular, ac_simplices_singular]]
    assert lengths == [3, 16, 0, 0, 3]

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='all')]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='exterior')]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='interior')]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='regular')]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='singular')]

    lengths = [len(sim) for sim in [
        triangles_all, triangles_exterior, triangles_interior, triangles_regular, triangles_singular]]
    assert lengths == [0, 11, 0, 0, 0]

    H = alpha_complex.graph_from_lines(alpha, type='all')
    assert isinstance(H, nx.Graph)
    H = alpha_complex.graph_from_triangles(alpha, type='regular')
    assert isinstance(H, nx.Graph)


def test_AlphaComplex_2():
    points = np.array([[10, 10], [10, 14], [11, 11], [12, 11], [13, 10], [13, 13], [15, 14], [17, 16],
                       [20, 20], [20, 24], [21, 21], [22, 21], [23, 20], [23, 23], [25, 24],
                       [10, 24]])
    assert len(points) == 16
    alpha_complex = AlphaComplex(points)
    assert len(alpha_complex.lines) == 38
    assert alpha_complex.dimension == 2
    assert alpha_complex.get_alpha_complex_lines(alpha=alpha_complex.optimal_alpha(), type='exterior') == []
    assert alpha_complex.optimal_alpha() in alpha_complex.alphas()
    assert np.max(alpha_complex.alphas()) == np.inf
    assert isinstance(alpha_complex.alpha_shape(alpha=2.2), AlphaShape)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type='all')
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(alpha, type='exterior')
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(alpha, type='interior')
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type='regular')
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(alpha, type='singular')

    lengths = [len(sim) for sim in [
        ac_simplices_all, ac_simplices_exterior, ac_simplices_interior, ac_simplices_regular, ac_simplices_singular]]
    assert lengths == [25, 13, 10, 10, 5]

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='all')]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='exterior')]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='interior')]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='regular')]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='singular')]

    lengths = [len(sim) for sim in [
        triangles_all, triangles_exterior, triangles_interior, triangles_regular, triangles_singular]]
    assert lengths == [12, 11, 3, 7, 2]

    H = alpha_complex.graph_from_lines(alpha, type='regular')
    assert isinstance(H, nx.Graph)
    H = alpha_complex.graph_from_triangles(alpha, type='regular')
    assert isinstance(H, nx.Graph)


def test_AlphaComplex_3():
    points = np.array([[10, 10], [10, 20], [20, 20], [20, 10], [10, 15], [15, 20], [20, 15], [15, 10],
                       [10, 12], [10, 18], [20, 18], [20, 12], [17, 20], [12, 10], [17, 10], [12, 20],
                       [11, 11], [11, 19], [19, 19], [19, 11], [11, 15], [15, 19], [19, 15], [15, 11],
                       [13, 11], [11.5, 12], [12, 18], [13, 19], [17, 19], [18, 18], [17, 11], [18, 12],
                       [15, 15]
                       ])
    assert len(points) == 33
    alpha_complex = AlphaComplex(points)
    assert len(alpha_complex.lines) == 80
    assert alpha_complex.dimension == 2
    assert alpha_complex.get_alpha_complex_lines(alpha=alpha_complex.optimal_alpha(), type='exterior') == []
    assert alpha_complex.optimal_alpha() in alpha_complex.alphas()
    assert np.max(alpha_complex.alphas()) == np.inf
    assert isinstance(alpha_complex.alpha_shape(alpha=2.2), AlphaShape)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type='all')
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(alpha, type='exterior')
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(alpha, type='interior')
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type='regular')
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(alpha, type='singular')

    lengths = [len(sim) for sim in [
        ac_simplices_all, ac_simplices_exterior, ac_simplices_interior, ac_simplices_regular, ac_simplices_singular]]
    assert lengths == [75, 5, 40, 28, 7]

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='all')]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='exterior')]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='interior')]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='regular')]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type='singular')]

    lengths = [len(sim) for sim in [
        triangles_all, triangles_exterior, triangles_interior, triangles_regular, triangles_singular]]
    assert lengths == [39, 9, 8, 28, 3]

    H = alpha_complex.graph_from_lines(alpha=2.2, type='regular')
    assert isinstance(H, nx.Graph)
    H = alpha_complex.graph_from_triangles(alpha, type='regular')
    assert isinstance(H, nx.Graph)


def test_AlphaShape_0():
    points = np.array([])

    alpha = 2.2
    alpha_shape = AlphaShape(alpha, points=points)
    assert len(alpha_shape.alpha_shape) == 0
    assert isinstance(alpha_shape.alpha_shape, list)
    assert alpha_shape.dimension is None
    assert isinstance(alpha_shape.region, EmptyRegion)
    assert alpha_shape.region_measure == 0
    assert alpha_shape.subregion_measure == 0
    assert alpha_shape.n_points_alpha_shape == 0
    assert math.isnan(alpha_shape.n_points_alpha_shape_rel)
    assert alpha_shape.n_points_on_boundary == 0
    assert math.isnan(alpha_shape.n_points_on_boundary_rel)
    assert alpha_shape.vertex_indices == []
    assert np.size(alpha_shape.vertices) == 0
    assert np.size(alpha_shape.vertices_alpha_shape) == 0
    assert np.size(alpha_shape.vertex_alpha_shape_indices) == 0

    cc = alpha_shape.connected_components
    assert len(cc) == 0
    assert alpha_shape.vertices_connected_components_indices == []


def test_AlphaShape_1():
    points = np.array([[10, 10], [10, 14], [11, 11], [12, 11], [13, 10], [13, 13], [15, 14], [17, 16],
                       [10, 24]])
    assert len(points) == 9

    alpha = 2.2
    alpha_shape = AlphaShape(alpha, points=points)
    assert len(alpha_shape.alpha_shape) == 13
    assert isinstance(alpha_shape.alpha_shape, list)
    assert alpha_shape.dimension == 2
    assert isinstance(alpha_shape.region, Region2D)
    assert alpha_shape.region_measure == 10.5
    assert alpha_shape.subregion_measure == 13.16227766016838
    assert alpha_shape.n_points_alpha_shape == 8
    assert alpha_shape.n_points_alpha_shape_rel == 0.8888888888888888
    assert alpha_shape.n_points_on_boundary == 5
    assert alpha_shape.n_points_on_boundary_rel == 0.625
    assert np.array_equal(alpha_shape.vertex_indices, [0, 1, 2, 4, 5])
    assert alpha_shape.vertices.shape == (5, 2)
    assert alpha_shape.vertices_alpha_shape.shape == (8, 2)
    assert len(alpha_shape.vertex_alpha_shape_indices) == 8
    cc = alpha_shape.connected_components
    assert len(cc) == 1
    assert isinstance(cc[0], Region2D)
    assert alpha_shape.vertices_connected_components_indices == [[2, 1, 0, 5, 3, 4]]

    alpha = 1
    # alpha_shape = AlphaShape(alpha, points=points)
    alpha_shape.alpha = 1
    assert len(alpha_shape.alpha_shape) == 3
    assert isinstance(alpha_shape.alpha_shape, list)
    assert alpha_shape.dimension == 2
    assert isinstance(alpha_shape.region, EmptyRegion)
    assert alpha_shape.region_measure == 0
    assert alpha_shape.subregion_measure == 0
    assert alpha_shape.n_points_alpha_shape == 4
    assert alpha_shape.n_points_alpha_shape_rel == 0.4444444444444444
    assert alpha_shape.n_points_on_boundary == 0
    assert alpha_shape.n_points_on_boundary_rel == 0
    assert np.array_equal(alpha_shape.vertex_indices, [])
    assert alpha_shape.vertices.shape == (0, 2)
    assert alpha_shape.vertices_alpha_shape.shape == (4, 2)
    assert len(alpha_shape.vertex_alpha_shape_indices) == 4
    cc = alpha_shape.connected_components
    assert len(cc) == 0
    assert alpha_shape.vertices_connected_components_indices == []


@pytest.mark.visual
def test_AlphaShape_2_visual():
    points = np.array([[10, 10], [10, 14], [11, 11], [12, 11], [13, 10], [13, 13], [15, 14], [17, 16],
                       [20, 20], [20, 24], [21, 21], [22, 21], [23, 20], [23, 23], [25, 24],
                       [10, 24]])
    assert len(points) == 16

    alpha = 2.2
    alpha_shape = AlphaShape(alpha, points=points)
    H = alpha_shape.alpha_complex.graph_from_lines(alpha, type='regular')
    assert isinstance(H, nx.Graph)
    # visualization
    nx.draw_networkx(H, points, with_labels=True)
    plt.show()


def test_AlphaShape_2():
    points = np.array([[10, 10], [10, 14], [11, 11], [12, 11], [13, 10], [13, 13], [15, 14], [17, 16],
                       [20, 20], [20, 24], [21, 21], [22, 21], [23, 20], [23, 23], [25, 24],
                       [10, 24]])
    assert len(points) == 16

    alpha = 2.2
    alpha_shape = AlphaShape(alpha, points=points)
    H = alpha_shape.alpha_complex.graph_from_lines(alpha, type='regular')
    assert isinstance(H, nx.Graph)

    assert len(alpha_shape.alpha_shape) == 25
    assert isinstance(alpha_shape.alpha_shape, list)
    assert alpha_shape.dimension == 2
    assert isinstance(alpha_shape.region, Region2D)
    assert alpha_shape.region_measure == 21.0
    assert alpha_shape.subregion_measure == 26.32455532033676
    assert alpha_shape.n_points_alpha_shape == 15
    assert alpha_shape.n_points_alpha_shape_rel == 0.9375
    assert alpha_shape.n_points_on_boundary == 10
    assert alpha_shape.n_points_on_boundary_rel == 0.6666666666666666
    assert np.array_equal(alpha_shape.vertex_indices, [0, 1, 2, 4, 5, 8, 9, 10, 12, 13])
    assert alpha_shape.vertices.shape == (10, 2)
    assert alpha_shape.vertices_alpha_shape.shape == (15, 2)
    assert len(alpha_shape.vertex_alpha_shape_indices) == 15
    cc = alpha_shape.connected_components
    assert len(cc) == 2
    assert isinstance(cc[0], Region2D)
    assert alpha_shape.vertices_connected_components_indices == [[13, 11, 12, 8, 10, 9], [2, 1, 0, 5, 3, 4]]


def test_AlphaShape_3():
    points = np.array([[10, 10], [10, 20], [20, 20], [20, 10], [10, 15], [15, 20], [20, 15], [15, 10],
                       [10, 12], [10, 18], [20, 18], [20, 12], [17, 20], [12, 10], [17, 10], [12, 20],

                       [11, 11], [11, 19], [19, 19], [19, 11], [11, 15], [15, 19], [19, 15], [15, 11],
                       [13, 11], [11.5, 12], [12, 18], [13, 19], [17, 19], [18, 18], [17, 11], [18, 12],
                       [15, 15]
                       ])
    assert len(points) == 33

    alpha = 2.2
    alpha_shape = AlphaShape(alpha, points=points)
    H = alpha_shape.alpha_complex.graph_from_lines(alpha, type='regular')
    assert isinstance(H, nx.Graph)
    assert len(alpha_shape.alpha_shape) == 75
    assert isinstance(alpha_shape.alpha_shape, list)
    assert alpha_shape.dimension == 2
    assert isinstance(alpha_shape.region, Region2D)
    assert alpha_shape.region_measure == 65.0
    assert alpha_shape.subregion_measure == 73.81471965135825
    assert alpha_shape.n_points_alpha_shape == 33
    assert alpha_shape.n_points_alpha_shape_rel == 1.
    assert alpha_shape.n_points_on_boundary == 28
    assert alpha_shape.n_points_on_boundary_rel == 0.8484848484848485
    assert np.array_equal(alpha_shape.vertex_indices, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 20,
                                                       21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31])
    assert alpha_shape.vertices.shape == (28, 2)
    assert alpha_shape.vertices_alpha_shape.shape == (33, 2)
    assert len(alpha_shape.vertex_alpha_shape_indices) == 33
    cc = alpha_shape.connected_components
    assert len(cc) == 1
    assert isinstance(cc[0], Region2D)
    assert sorted(alpha_shape.vertices_connected_components_indices[0]) == sorted(
           [26, 20, 32, 9, 4, 8, 25, 24, 13, 7, 23, 31, 22, 29, 10, 6, 11, 15, 27, 5, 21, 17, 1, 16, 0, 18, 2, 12,
            28, 19, 3, 14, 30])



from pathlib import Path
import tempfile

from pandas.testing import assert_frame_equal

from locan.locan_io import save_asdf, load_asdf_file


def test_save_and_load_asdf(locdata_2d):
    # for visual inspection use:
    # io.save_asdf(locdata_2d, path=locan.ROOT_DIR / 'tests/test_data/locdata.asdf')
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / 'locdata.asdf'
        save_asdf(locdata_2d, path=file_path)

        locdata = load_asdf_file(path=file_path)
        # print(locdata.data)
        assert_frame_equal(locdata.data, locdata_2d.data)
        assert (locdata.meta.identifier == locdata_2d.meta.identifier)
        locdata_2d.properties.pop("localization_density_ch", None)
        locdata_2d.properties.pop("region_measure_ch", None)
        assert (locdata.properties == locdata_2d.properties)

        dat = load_asdf_file(path=file_path, nrows=5)
        assert (len(dat) == 5)



"""

Utility functions for file input/output of localization data.

"""
import io
from contextlib import closing
import logging

import pandas as pd

from locan.constants import PROPERTY_KEYS


__all__ = ['convert_property_types', 'convert_property_names']

logger = logging.getLogger(__name__)


def convert_property_types(dataframe, types, loc_properties=None):
    """
    Convert data types according to the column-type mapping in types.
    If the target type is one of 'integer', 'signed', 'unsigned', 'float'
    then :func:`pandas.to_numeric` will be applied.
    Otherwise, if the target type is any type object like `int`, `str`, `np.float64` or similar
    then :func:`pandas.astype` will be applied.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Data to be converted
    types : dict
        Mapping of loc_properties to types
    loc_properties : list(str), None
        The columns in dataframe to be converted. If None all columns will be converted according to types.

    Returns
    -------
    pandas.DataFrame
        A copy of dataframe with converted types
    """
    new_df = dataframe.copy()
    if loc_properties is None:
        loc_properties = dataframe.columns
    else:
        loc_properties = [key for key in loc_properties if key in dataframe.columns]
    selected_property_keys = {key: types[key] for key in loc_properties if key in types}

    for key, value in selected_property_keys.items():
        if isinstance(value, str):
            new_df[key] = pd.to_numeric(dataframe[key], downcast=value, errors='coerce')
        else:
            new_df[key] = dataframe[key].astype(value)

    return new_df


def open_path_or_file_like(path_or_file_like, mode='r', encoding=None):
    """
    Provide open-file context from `path_or_file_like` input.

    Parameters
    ----------
    path_or_file_like : str, bytes, os.PathLike, file-like
        Identifier for file
    mode
        same as in `open()`
    encoding
        same as in `open()`

    Returns
    -------
    context for file object
    """
    try:
        all(getattr(path_or_file_like, attr) for attr in ('seek', 'read', 'close'))
        file = path_or_file_like
    except (AttributeError, io.UnsupportedOperation):
        try:
            # if hasattr(path_or_file_like, "__fspath__")
            # or isinstance(path_or_file_like, (str, bytes)):
            file = open(path_or_file_like, mode=mode, encoding=encoding)
        except TypeError:
            raise TypeError("path_or_file_like must be str, bytes, os.PathLike or file-like.")
    return closing(file)


def convert_property_names(properties, property_mapping=None):
    """
    Convert property names to standard locan property names if a mapping is provided.
    Otherwise leave the property name a is and throw a warning.

    Parameters
    ----------
    properties : list[str], tuple[str]
        Properties to be converted
    property_mapping : dict[str: str], list[dict]
        Mappings between other property names and locan property names

    Returns
    -------
    list[str]
        Converted property names
    """
    if property_mapping is None:
        property_mapping_ = {}
    elif isinstance(property_mapping, (list, tuple)):
        property_mapping_ = {}
        for mapping in property_mapping:
            property_mapping_.update(mapping)
    else:
        property_mapping_ = property_mapping

    column_keys = []
    for i in properties:
        if i in PROPERTY_KEYS.keys():
            column_keys.append(i)
        elif i in property_mapping_:
            column_keys.append(property_mapping_[i])

        else:
            logger.warning(f'Column {i} is not a Locan property standard.')
            column_keys.append(i)

    return column_keys



import pytest
from locan.datasets import load_npc, load_tubulin
from locan.dependencies import HAS_DEPENDENCY

# pytestmark = pytest.mark.requires_datasets
pytestmark = pytest.mark.skipif(not HAS_DEPENDENCY["requests"], reason="requires requests for html download")


def test_load_npc():
    locdata = load_npc()
    assert locdata.meta.element_count == 2285189


def test_load_tubulin():
    locdata = load_tubulin()
    assert locdata.meta.element_count == 1506568



"""

Configuration variables used throughout the project.

.. autosummary::
   :toctree: ./

   DATASETS_DIR
   RENDER_ENGINE
   N_JOBS
   COLORMAP_CONTINUOUS
   COLORMAP_DIVERGING
   COLORMAP_CATEGORICAL
   TQDM_LEAVE
   TQDM_DISABLE
"""
from __future__ import annotations

from pathlib import Path

import matplotlib.colors as mcolors

from locan.dependencies import HAS_DEPENDENCY
if HAS_DEPENDENCY["colorcet"]: from colorcet import m_fire, m_gray, m_coolwarm, m_glasbey_dark
from locan.constants import RenderEngine

__all__ = ['DATASETS_DIR', 'RENDER_ENGINE', 'N_JOBS', 'TQDM_LEAVE', 'TQDM_DISABLE',
           'COLORMAP_CONTINUOUS', 'COLORMAP_DIVERGING', 'COLORMAP_CATEGORICAL'
           ]


#: Standard directory for example datasets.
DATASETS_DIR = Path.home() / 'LocanDatasets'

#: Render engine.
RENDER_ENGINE = RenderEngine.MPL

#: The number of cores that are used in parallel for some algorithms.
#: Following the scikit convention: n_jobs is the number of parallel jobs to run.
#: If -1, then the number of jobs is set to the number of CPU cores.
N_JOBS: int = 1

#: Leave tqdm progress bars after finishing the iteration.
#: Flag to leave tqdm progress bars.
TQDM_LEAVE: bool = True

#: Disable tqdm progress bars.
#: Flag to disable all tqdm progress bars.
TQDM_DISABLE: bool = False

#: Default colormap for continuous scales. Default is `colorcet.m_fire` if installed or 'viridis'.
COLORMAP_CONTINUOUS: mcolors.Colormap | str = 'viridis'

#: Default colormap for diverging scales. Default is `colorcet.m_coolwarm` if installed or 'coolwarm'.
COLORMAP_DIVERGING: mcolors.Colormap | str = 'coolwarm'

#: Default colormap for categorical scales. Default is `colorcet.m_glasbey_dark` if installed or 'tab20'.
COLORMAP_CATEGORICAL: mcolors.Colormap | str = 'tab20'

if HAS_DEPENDENCY["colorcet"]:
    COLORMAP_CONTINUOUS = m_fire
    COLORMAP_DIVERGING = m_coolwarm
    COLORMAP_CATEGORICAL = m_glasbey_dark



#!/usr/bin/env python

"""
Show original SMLM images overlaid with localization data.
Data is rendered in napari.

To run the script::

    locan check <pixel size> -f <images file> -l <localization file> -t <file type>

Try for instance::

    locan check 133 -f "locan/tests/test_data/images.tif" -l "locan/tests/test_data/rapidStorm_from_images.txt" -t 2
"""
import argparse
from pathlib import Path

import numpy as np
import tifffile as tif

from locan.dependencies import HAS_DEPENDENCY
import locan as lc

if HAS_DEPENDENCY["napari"]: import napari


def render_locs_per_frame_napari(images, pixel_size, locdata, viewer=None, transpose=True,
                                 kwargs_image=None, kwargs_points=None):
    """
    Display original recording and overlay localization spots in napari.

    Parameters
    ---------
    images : np.array
        Stack of raw data as recorded by camera.
    pixel_size : float, tuple of float with shape (2,)
        Pixel size for images (in locdata units).
    transpose : bool
        If True transpose x and y axis of `images`.
    locdata : LocData
        Localization data that corresponds to `images` raw data.
    viewer : napari viewer
        The viewer object on which to add the image
    kwargs_image : dict
        Other parameters passed to napari.Viewer().add_image().
    kwargs_points : dict
        Other parameters passed to napari.Viewer().add_points().

    Returns
    -------
    napari Viewer object
        Viewer with the image.
    """
    if kwargs_image is None:
        kwargs_image = {}
    if kwargs_points is None:
        kwargs_points = {}

    if np.ndim(pixel_size) == 0:
        pixel_size_ = (pixel_size, pixel_size)
    elif np.ndim(pixel_size) == 1 and len(pixel_size) == 2:
        pixel_size_ = pixel_size
    else:
        raise TypeError('Dimension of `pixel_size` is incompatible with 2d image.')

    # transpose x and y axis
    if transpose:
        images_ = np.transpose(images, (0, 2, 1))
    else:
        images_ = images

    points = locdata.data[locdata.data['frame']<len(images)][['frame', 'position_x', 'position_y']].values

    # Provide napari viewer if not provided
    if viewer is None:
        viewer = napari.Viewer()

    viewer.add_image(images_, name=f'Raw data', **kwargs_image, scale=pixel_size_)
    viewer.add_points(data=points, name=f'LocData {lc.locdata_id}',
                      symbol='disc', size=500, face_color='r', edge_color='r', opacity=0.3,
                      **kwargs_points)

    return viewer


def sc_check(pixel_size, file_images=None, file_locdata=None, file_type=lc.FileType.RAPIDSTORM,
             viewer=None, transpose=True, kwargs_image=None, kwargs_points=None):
    """
    Load and display original recording and load and overlay localization spots in napari.

    Parameters
    ---------
    pixel_size : float or tuple of float with shape (2,)
        Pixel size for images (in locdata units).
    file_images : str, bytes, os.PathLike
        File path for stack of raw data as recorded by camera.
    file_locdata : str, bytes, os.PathLike
        File path for localization data that corresponds to `images` raw data.
    file_type : int, str, locan.constants.FileType, locan.data.metadata_pb2.Metadata
        Indicator for the file type.
        Integer or string should be according to locan.constants.FileType.
    transpose : bool
        If True transpose x and y axis of `images`.
    locdata : LocData
        Localization data that corresponds to `images` raw data.
    viewer : napari.Viewer
        The viewer object on which to add the image
    kwargs_image : dict
        Other parameters passed to :meth:`napari.Viewer.add_image`.
    kwargs_points : dict
        Other parameters passed to :meth:`napari.Viewer.add_points`.

    Returns
    -------
    napar.Viewer
        Viewer with the image.
    """
    if kwargs_image is None:
        kwargs_image = {}
    if kwargs_points is None:
        kwargs_points = {}

    with napari.gui_qt():
        # load images
        if file_images is None:
            file_images = lc.file_dialog(directory=None,
                                         message='Select images file...', filter='Tif files (*.tif)')[0]
        # load images from tiff file
        image_stack = tif.imread(str(file_images))  #, key=range(0, 10, 1))  - maybe add key parameter

        # load locdata
        if file_locdata is None:
            file_locdata = lc.file_dialog(directory=None,
                                          message='Select a localization file...',
                                          filter='Text files (*.txt);; CSV files (*.csv)')[0]
        locdata = lc.load_locdata(file_locdata, file_type=file_type)

    # due to changed napari behavior from v3.0 on the context manager is moved up.
    # with napari.gui_qt():
        render_locs_per_frame_napari(image_stack, pixel_size, locdata, viewer, transpose,
                                     kwargs_image, kwargs_points)


def _add_arguments(parser):
    parser.add_argument(dest='pixel_size', type=float,
                        help='Pixel size for images (in locdata units).')
    parser.add_argument('-f', '--file', dest='file_images', type=str, default=None,
                        help='File with images of original recording.')
    parser.add_argument('-l', '--localizations', dest='file_locdata', type=str, default=None,
                        help='File with localization data.')
    parser.add_argument('-t', '--type', dest='file_type', type=int, default=2,
                        help='Integer or string indicating the file type.')


def main(args=None):

    parser = argparse.ArgumentParser(description='Show localizations in original recording.')
    _add_arguments(parser)
    returned_args = parser.parse_args(args)

    sc_check(pixel_size=returned_args.pixel_size, file_images=returned_args.file_images,
             file_locdata=returned_args.file_locdata, file_type=returned_args.file_type,
             transpose=True, kwargs_image={}, kwargs_points={})


if __name__ == '__main__':
    main()



"""
Simulate drift and apply it to localization data.

Drift can be linear in time or resemble a random walk.
"""
import sys

import numpy as np
import pandas as pd

from locan.data.locdata import LocData
from locan.data.metadata_utils import _modify_meta


__all__ = ['add_drift']


def _random_walk_drift(n_steps, diffusion_constant, velocity, seed=None):
    """
    Transform locdata coordinates according to a simulated drift.

    Position deltas are computed as function of frame number.
    Within a single time unit delta_t the probability for a diffusion step of length delta_x is:

    p(delta_x, delta_t) = Norm(-velocity*delta_t, sigma**2) where the standard deviation sigma**2 = 2 * D * delta_t

    Parameters
    ----------
    n_steps : int
        The number of time steps for the random walk.
    diffusion_constant : tuple of float with shape (point_dimension,)
        Diffusion constant for each dimension specifying the drift velocity.
        The diffusion constant has the unit of square of localization coordinate unit per frame unit.
    velocity : tuple of float with shape (point_dimension,)
        Drift velocity in units of localization coordinate unit per frame unit
    seed : int
        random number generation seed

    Returns
    -------
    numpy.ndarray with shape (n_points, diffusion_constant.shape)
        Position deltas that have to be added to the original localizations.
    """
    if seed is not None:
        np.random.seed(seed)
    sigmas = np.sqrt(2 * np.asarray(diffusion_constant))  # * delta_t which is 1 frame
    steps = np.array([np.random.normal(loc=-vel, scale=sigma, size=n_steps)
                      for vel, sigma in zip(velocity, sigmas)])
    return np.cumsum(steps, axis=1)


def _drift(frames, diffusion_constant=None, velocity=None, seed=None):
    """
    Compute position deltas as function of frame number.

     Parameters
    ----------
    frames : array-like of int
        array with frame numbers
    diffusion_constant : tuple of float with shape (point_dimension,), None
        Diffusion constant for each dimension specifying the drift velocity.
        The diffusion constant has the unit of square of localization coordinate unit per frame unit.
        If None only linear drift is computed.
    velocity : tuple of float with shape (point_dimension,)
        Drift velocity in units of localization coordinate unit per frame unit
    seed : int
        random number generation seed

    Returns
    -------
    numpy.ndarray with shape (n_points, diffusion_constant.shape)
        Position deltas that have to be added to the original localizations.
    """
    frames_ = np.asarray(frames)

    if diffusion_constant is None and velocity is not None:  # only linear drift
        position_deltas = frames_[:, np.newaxis] * velocity
        position_deltas = position_deltas.T

    elif diffusion_constant is not None:  # random walk plus linear drift
        velocity_ = (np.zeros(len(diffusion_constant)) if velocity is None else velocity)
        cumsteps = _random_walk_drift(frames_.max() + 1, diffusion_constant, velocity_, seed)
        position_deltas = cumsteps[:, frames_]

    else:  # no drift
        position_deltas = None

    return position_deltas


def add_drift(locdata, diffusion_constant=None, velocity=None, seed=None):
    """
    Compute position deltas as function of frame number.

    Parameters
    ----------
    locdata : LocData object
        Original localization data
    diffusion_constant : tuple of float with shape (point_dimension,), None
        Diffusion constant for each dimension specifying the drift velocity.
        The diffusion constant has the unit of square of localization coordinate unit per frame unit.
    velocity : tuple of float with shape (point_dimension,)
        Drift velocity in units of localization coordinate unit per frame unit
    seed : int
        random number generation seed

    Returns
    -------
    LocData
        A new LocData instance with localization data.
    """
    local_parameter = locals()

    if len(locdata) == 0:
        return locdata

    points = locdata.coordinates
    frames = locdata.data['frame']

    if diffusion_constant is None and velocity is None:
        transformed_points = points
    else:
        transformed_points = points + _drift(frames, diffusion_constant, velocity, seed).T

    # new LocData object
    new_dataframe = locdata.data.copy()
    new_dataframe.update(pd.DataFrame(transformed_points, columns=locdata.coordinate_labels,
                                      index=locdata.data.index))
    new_locdata = LocData.from_dataframe(new_dataframe)

    # update metadata
    meta_ = _modify_meta(locdata, new_locdata, function_name=sys._getframe().f_code.co_name,
                         parameter=local_parameter,
                         meta=None)
    new_locdata.meta = meta_

    return new_locdata



"""
This module provides functions for clustering localizations.

The functions take LocData as input and compute new LocData objects representing collections of clustered
localizations.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   clustering
   utils

"""

from locan.data.cluster.clustering import *
from locan.data.cluster.utils import *

__all__ = []
__all__.extend(clustering.__all__)
__all__.extend(utils.__all__)



import pytest
import numpy as np
import matplotlib.pyplot as plt
from shapely.geometry import MultiPoint
from shapely import affinity

from locan import BoundingBox, ConvexHull, OrientedBoundingBox
from locan.data.hulls.hull import _ConvexHullScipy, _ConvexHullShapely


def test_BoundingBox_2d(locdata_2d):
    true_hull = np.array([[1, 1], [5, 6]])
    hull = BoundingBox(locdata_2d.coordinates)
    assert hull.width.shape == (2,)
    assert(locdata_2d.properties['region_measure_bb'] == hull.region_measure == 20)
    assert(hull.subregion_measure == 18)
    np.testing.assert_array_equal(hull.hull, true_hull)
    assert hull.region.region_measure == 20


def test_BoundingBox_3d(locdata_3d):
    true_hull = np.array([[1, 1, 1], [5, 6, 5]])
    hull = BoundingBox(locdata_3d.coordinates)
    assert hull.width.shape == (3,)
    assert(hull.region_measure == 80)
    assert(hull.subregion_measure == 26)
    np.testing.assert_array_equal(hull.hull, true_hull)
    with pytest.raises(NotImplementedError):
        assert hull.region.region_measure == 80


@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', 0),
    ('locdata_single_localization', 0),
    ('locdata_non_standard_index', 20)
])
def test_BoundingBox(locdata_empty, locdata_single_localization, locdata_non_standard_index,
                     fixture_name, expected):
    locdata = eval(fixture_name)
    hull = BoundingBox(locdata.coordinates)
    assert hull.region_measure == expected


def test_ConvexHullScipy(locdata_2d):
    true_convex_hull_indices = np.array([5, 3, 1, 0, 4])
    hull = _ConvexHullScipy(locdata_2d.coordinates)
    assert np.array_equal(hull.vertex_indices, true_convex_hull_indices)
    assert np.array_equal(hull.vertices, locdata_2d.coordinates[true_convex_hull_indices])
    assert hull.points_on_boundary == 5
    assert hull.region_measure == 14
    assert hull.region.region_measure == 14


def test_ConvexHullShapely(locdata_2d):
    true_convex_hull_indices = np.array([0, 1, 3, 5, 4])
    hull = _ConvexHullShapely(locdata_2d.coordinates)
    # assert np.array_equal(hull.vertex_indices, true_convex_hull_indices)
    assert np.array_equal(hull.vertices, locdata_2d.coordinates[true_convex_hull_indices])
    assert hull.points_on_boundary == 5
    assert hull.region_measure == 14
    assert hull.region.region_measure == 14


def test_ConvexHull_2d(locdata_2d):
    true_convex_hull_indices = np.array([5, 3, 1, 0, 4])
    hull = ConvexHull(locdata_2d.coordinates, method='scipy')
    assert np.array_equal(hull.vertex_indices, true_convex_hull_indices)
    assert np.array_equal(hull.vertices, locdata_2d.coordinates[true_convex_hull_indices])
    assert hull.points_on_boundary == 5
    assert hull.region_measure == 14
    assert hull.region.region_measure == 14

    true_convex_hull_indices = np.array([0, 1, 3, 5, 4])
    hull = ConvexHull(locdata_2d.coordinates, method='shapely')
    # assert np.array_equal(hull.vertex_indices, true_convex_hull_indices)
    assert np.array_equal(hull.vertices, locdata_2d.coordinates[true_convex_hull_indices])
    assert hull.points_on_boundary == 5
    assert hull.region_measure == 14
    assert hull.region.region_measure == 14


def test_ConvexHull_3d(locdata_3d):
    true_convex_hull_indices = np.array([0, 1, 2, 3, 4, 5])
    hull = ConvexHull(locdata_3d.coordinates, method='scipy')
    assert np.array_equal(hull.vertex_indices, true_convex_hull_indices)
    assert np.array_equal(hull.vertices, locdata_3d.coordinates[true_convex_hull_indices])
    assert hull.points_on_boundary == 6
    assert hull.region_measure == 19.833333333333332
    with pytest.raises(NotImplementedError):
        assert hull.region.region_measure == 19.833333333333332

    with pytest.raises(TypeError):
        ConvexHull(locdata_3d.coordinates, method='shapely')


@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', 0),
    ('locdata_single_localization', 0)
])
def test_ConvexHull(locdata_empty, locdata_single_localization,
                    fixture_name, expected):
    locdata = eval(fixture_name)
    with pytest.raises(TypeError):
        ConvexHull(locdata.coordinates, method='scipy')
    with pytest.raises(TypeError):
        ConvexHull(locdata.coordinates, method='shapely')


@pytest.mark.visual
def test_OrientedBoundingBox_2d_points_visual():
    points = np.array([[0, 0], [0, 2], [1, 2], [1, 0], [0, 0]])

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(5, 5))
    for angle in np.linspace(0, 180, 5):
        rotated_points = affinity.rotate(MultiPoint(points), angle, origin=[0, 0], use_radians=False)
        rotated_points = np.array([rpts.coords[0] for rpts in rotated_points.geoms])
        ax.plot(*rotated_points.T)

        obb = OrientedBoundingBox(rotated_points)
        assert int(obb.angle) in [0, 45, 90, -45]
        print(obb.region)
        print(obb.vertices)
        print(obb.width)
        print(angle, obb.angle)
        ax.plot(*obb.vertices.T)
        ax.add_patch(obb.region.as_artist())
    ax.set(xlim=(-3, 3), ylim=(-3, 3))
    plt.show()


def test_OrientedBoundingBox_2d_points():
    points = np.array([[0, 0], [0, 2], [1, 2], [1, 0], [0, 0]])
    for angle in np.linspace(0, 180, 5):
        rotated_points = affinity.rotate(MultiPoint(points), angle, origin=[0, 0], use_radians=False)
        rotated_points = np.array([rpts.coords[0] for rpts in rotated_points.geoms])
        obb = OrientedBoundingBox(rotated_points)
        assert int(obb.angle) in [0, 45, 90, -45, 135]
        assert np.isclose(obb.region_measure, 2)
        assert np.isclose(obb.subregion_measure, 6)


def test_OrientedBoundingBox_2d(locdata_2d):
    hull = OrientedBoundingBox(locdata_2d.coordinates)
    assert hull.width.shape == (2,)
    assert(locdata_2d.properties['region_measure_bb'] == hull.region_measure == 20)
    assert(hull.subregion_measure == 18)
    assert hull.region.region_measure == 20


def test_OrientedBoundingBox_3d(locdata_3d):
    with pytest.raises(TypeError):
        OrientedBoundingBox(locdata_3d.coordinates)


@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', 0),
    ('locdata_single_localization', 0),
    ('locdata_non_standard_index', 20)
])
def test_OrientedBoundingBox(locdata_empty, locdata_single_localization, locdata_non_standard_index,
                             fixture_name, expected):
    locdata = eval(fixture_name)
    hull = OrientedBoundingBox(locdata.coordinates)
    assert hull.region_measure == expected



"""

Register localization data.

This module registers localization data and provides transformation parameters to put other localization data
in registry.

Parts of this code is adapted from https://github.com/jungmannlab/picasso.
(MIT license, Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry)
"""
from collections import namedtuple

import numpy as np
from lmfit import Model, Parameters
import matplotlib.pyplot as plt

from locan.data.locdata import LocData
from locan.data.properties.locdata_statistics import range_from_collection
from locan.data.transform.transformation import _homogeneous_matrix
from locan.data.aggregate import histogram
from locan.dependencies import HAS_DEPENDENCY, needs_package

if HAS_DEPENDENCY["open3d"]: import open3d as o3d


__all__ = ['register_icp', 'register_cc']


@needs_package("open3d")
def _register_icp_open3d(points, other_points, matrix=None, offset=None, pre_translation=None,
                         max_correspondence_distance=1_000, max_iteration=10_000, with_scaling=True, verbose=True):
    """
    Register `points` by an "Iterative Closest Point" algorithm using open3d.

    Parameters
    ----------
    points : array-like
        Points representing the source on which to perform the manipulation.
    other_points : array-like
        Points representing the target.
    matrix : tuple with shape (d, d)
        Transformation matrix used as initial value. If None the unit matrix is used.
    offset : tuple of int or float with shape (d,)
        Translation vector used as initial value. If None a vector of zeros is used.
    pre_translation : tuple of int or float
        Values for translation of coordinates before registration.
    max_correspondence_distance : float
        Threshold distance for the icp algorithm. Parameter is passed to open3d algorithm.
    max_iteration : int
        Maximum number of iterations. Parameter is passed to open3d algorithm.
    with_scaling : bool
        Allow scaling transformation. Parameter is passed to open3d algorithm.
    verbose : bool
        Flag indicating if transformation results are printed out.

    Returns
    -------
    namedtuple('Transformation', 'matrix offset')
        Matrix and offset representing the optimized transformation.
    """
    points_ = np.asarray(points)
    other_points_ = np.asarray(other_points)

    # prepare 3d
    if np.shape(points)[1] == np.shape(other_points)[1]:
        dimension = np.shape(points)[1]
    else:
        raise ValueError('Dimensions for locdata and other_locdata are incompatible.')

    if dimension == 2:
        points_3d = np.concatenate([points_, np.zeros((len(points_), 1))], axis=1)
        other_points_3d = np.concatenate([other_points_, np.zeros((len(other_points_), 1))], axis=1)
    elif dimension == 3:
        points_3d = points_
        other_points_3d = other_points_
    else:
        raise ValueError('Point array has the wrong shape.')

    # points in open3d
    point_cloud = o3d.geometry.PointCloud()
    point_cloud.points = o3d.utility.Vector3dVector(points_3d)

    other_point_cloud = o3d.geometry.PointCloud()
    other_point_cloud.points = o3d.utility.Vector3dVector(other_points_3d)

    # initial matrix
    matrix_ = np.identity(dimension) if matrix is None else matrix
    offset_ = np.zeros(dimension) if offset is None else offset

    matrix_3d = np.identity(3)
    matrix_3d[:dimension, :dimension] = matrix_
    offset_3d = np.zeros(3)
    offset_3d[:dimension] = offset_
    matrix_homogeneous = _homogeneous_matrix(matrix_3d, offset_3d)

    if pre_translation is not None:
        pre_translation_3d = np.zeros(3)
        pre_translation_3d[:dimension] = pre_translation
        other_point_cloud.translate(pre_translation_3d)

    # apply ICP
    registration = o3d.pipelines.registration.registration_icp(
        source=point_cloud, target=other_point_cloud,
        max_correspondence_distance=max_correspondence_distance,
        init=matrix_homogeneous,
        estimation_method=o3d.pipelines.registration.TransformationEstimationPointToPoint(with_scaling=with_scaling),
        criteria=o3d.pipelines.registration.ICPConvergenceCriteria(max_iteration=max_iteration)
    )

    if dimension == 2:
        new_matrix = registration.transformation[0:2, 0:2]
        new_offset = registration.transformation[0:2, 3]
    else:  # if dimension == 3:
        new_matrix = registration.transformation[0:3, 0:3]
        new_offset = registration.transformation[0:3, 3]

    if verbose:
        print(registration)

    Transformation = namedtuple('Transformation', 'matrix offset')
    return Transformation(new_matrix, new_offset)


def register_icp(locdata, other_locdata, matrix=None, offset=None, pre_translation=None,
                 max_correspondence_distance=1_000, max_iteration=10_000, verbose=True):
    """
    Register `points` or coordinates in `locdata` by an "Iterative Closest Point" algorithm using open3d.

    Parameters
    ----------
    locdata : array-like or LocData
        Localization data representing the source on which to perform the manipulation.
    other_locdata : array-like or LocData
        Localization data representing the target.
    matrix : tuple with shape (d, d)
        Transformation matrix used as initial value. If None the unit matrix is used.
    offset : tuple of int or float with shape (d,)
        Translation vector used as initial value. If None a vector of zeros is used.
    pre_translation : tuple of int or float
        Values for translation of coordinates before registration.
    max_correspondence_distance : float
        Threshold distance for the icp algorithm. Parameter is passed to open3d algorithm.
    max_iteration : int
        Maximum number of iterations. Parameter is passed to open3d algorithm.
    verbose : bool
        Flag indicating if transformation results are printed out.

    Returns
    -------
    namedtuple('Transformation', 'matrix offset')
        Matrix and offset representing the optimized transformation.
    """
    local_parameter = locals()

    # adjust input
    if isinstance(locdata, LocData):
        points = locdata.coordinates
    else:
        points = locdata

    if isinstance(other_locdata, LocData):
        other_points = other_locdata.coordinates
    else:
        other_points = other_locdata

    transformation = _register_icp_open3d(points, other_points, matrix=matrix, offset=offset,
                                                  pre_translation=pre_translation,
                                                  max_correspondence_distance=1_000, max_iteration=10_000,
                                                  verbose=True)
    return transformation


def _xcorr(imageA, imageB):
    """
    This function is adapted from picasso/imageprocess by Joerg Schnitzbauer, MPI of Biochemistry
    https://github.com/jungmannlab/picasso/blob/master/picasso/imageprocess.py
    (MIT license, Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry)
    """
    FimageA = np.fft.fft2(imageA)
    CFimageB = np.conj(np.fft.fft2(imageB))
    return np.fft.fftshift(
        np.real(np.fft.ifft2((FimageA * CFimageB)))
        ) / np.sqrt(imageA.size)


def _get_image_shift(imageA, imageB, box, roi=None, display=False):
    """
    Computes the shift from imageA to imageB.

    This function is adapted from picasso/imageprocess by Joerg Schnitzbauer, MPI of Biochemistry
    https://github.com/jungmannlab/picasso/blob/master/picasso/imageprocess.py
    (MIT license, Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry)
    """
    if (np.sum(imageA) == 0) or (np.sum(imageB) == 0):
        return 0, 0
    # Compute image correlation
    XCorr = _xcorr(imageA, imageB)
    # Cut out center roi
    Y, X = imageA.shape
    if roi is not None:
        Y_ = int((Y - roi) / 2)
        X_ = int((X - roi) / 2)
        if Y_ > 0:
            XCorr = XCorr[Y_:-Y_, :]
        else:
            Y_ = 0
        if X_ > 0:
            XCorr = XCorr[:, X_:-X_]
        else:
            X_ = 0
    else:
        Y_ = X_ = 0
    # A quarter of the fit ROI
    fit_X = int(box / 2)
    # A coordinate grid for the fitting ROI
    y, x = np.mgrid[-fit_X: fit_X + 1, -fit_X: fit_X + 1]
    # Find the brightest pixel and cut out the fit ROI
    y_max_, x_max_ = np.unravel_index(XCorr.argmax(), XCorr.shape)
    FitROI = XCorr[
        y_max_ - fit_X: y_max_ + fit_X + 1,
        x_max_ - fit_X: x_max_ + fit_X + 1,
    ]

    dimensions = FitROI.shape

    if 0 in dimensions or dimensions[0] != dimensions[1]:
        xc, yc = 0, 0
    else:
        # The fit model based on lmfit
        def flat_2d_gaussian(a, xc, yc, s, b):
            A = a * np.exp(-0.5 * ((x - xc) ** 2 + (y - yc) ** 2) / s ** 2) + b
            return A.flatten()

        gaussian2d = Model(
            flat_2d_gaussian, name="2D Gaussian", independent_vars=[]
        )

        # Set up initial parameters and fit
        params = Parameters()
        params.add("a", value=FitROI.max(), vary=True, min=0)
        params.add("xc", value=0, vary=True)
        params.add("yc", value=0, vary=True)
        params.add("s", value=1, vary=True, min=0)
        params.add("b", value=FitROI.min(), vary=True, min=0)
        results = gaussian2d.fit(FitROI.flatten(), params)

        # Get maximum coordinates and add offsets
        xc = results.best_values["xc"]
        yc = results.best_values["yc"]
        xc += X_ + x_max_
        yc += Y_ + y_max_

        if display:
            plt.figure(figsize=(17, 10))
            plt.subplot(1, 3, 1)
            plt.imshow(imageA, interpolation="none")
            plt.subplot(1, 3, 2)
            plt.imshow(imageB, interpolation="none")
            plt.subplot(1, 3, 3)
            plt.imshow(XCorr, interpolation="none")
            plt.plot(xc, yc, "x")
            plt.show()

        xc -= np.floor(X / 2)
        yc -= np.floor(Y / 2)

    return -yc, -xc


def register_cc(locdata, other_locdata, max_offset=None,
                bins=None, n_bins=None, bin_size=None, bin_edges=None, bin_range=None,
                verbose=False, **kwargs):
    """
    Register `points` or coordinates in `locdata` by a cross-correlation algorithm.

    This function is based on code from picasso/imageprocess by Joerg Schnitzbauer, MPI of Biochemistry
    https://github.com/jungmannlab/picasso/blob/master/picasso/imageprocess.py

    Parameters
    ----------
    locdata : array-like or LocData
        Localization data representing the source on which to perform the manipulation.
    other_locdata : array-like or LocData
        Localization data representing the target.
    max_offset : int or float or None
        Maximum possible offset.
    bins : int or sequence or `Bins` or `boost_histogram.axis.Axis` or None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple, list, numpy.ndarray of float with shape (dimension, n_bin_edges) or None
        Array of bin edges for all or each dimension.
    n_bins : int, list, tuple or numpy.ndarray or None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float, list, tuple or numpy.ndarray or None
        The size of bins in units of locdata coordinate units for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple or tuple of tuples of float with shape (dimension, 2) or None or 'zero'
        The data bin_range to be taken into consideration for all or each dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    verbose : bool
        Flag indicating if transformation results are printed out.

    Returns
    -------
    namedtuple('Transformation', 'matrix offset')
        Matrix and offset representing the optimized transformation.
    """
    if isinstance(locdata, LocData) and isinstance(other_locdata, LocData):
        if bin_range is None:
            bin_range_ = range_from_collection([locdata, other_locdata])
        else:
            bin_range_ = bin_range

        image, bins_, labels_ = histogram(locdata, bins=bins, n_bins=n_bins, bin_size=bin_size, bin_edges=bin_edges,
                                        bin_range=bin_range_)
        other_image, _, _ = histogram(other_locdata, bins=bins, n_bins=n_bins, bin_size=bin_size, bin_edges=bin_edges,
                                      bin_range=bin_range_)

    else:
        image = np.asarray(locdata)
        other_image = np.asarray(other_locdata)

    dimension = image.ndim
    matrix = np.identity(dimension)
    # todo: turn box into parameter
    offset = _get_image_shift(image, other_image, box=5, roi=max_offset, display=verbose)
    offset = tuple(np.asarray(offset) * bin_size)

    Transformation = namedtuple('Transformation', 'matrix offset')
    return Transformation(matrix, offset)



"""

Transform localization data with a BUnwarpJ transformation matrix.

This module provides functions to transform coordinates in LocData objects by applying a B-spline transformation as
defined with the ImageJ/Fiji plugin BunwarpJ_ [1]_, [2]_.

.. _BunwarpJ: https://imagej.net/BUnwarpJ

References
----------

.. [1] I. Arganda-Carreras, C. O. S. Sorzano, R. Marabini, J.-M. Carazo, C. Ortiz-de Solorzano, and J. Kybic,
   "Consistent and Elastic Registration of Histological Sections using Vector-Spline Regularization",
   Lecture Notes in Computer Science, Springer Berlin / Heidelberg, volume 4241/2006,
   CVAMIA: Computer Vision Approaches to Medical Image Analysis, pages 85-95, 2006.

.. [2] C.Ó. Sánchez Sorzano, P. Thévenaz, M. Unser,
   "Elastic Registration of Biological Images Using Vector-Spline Regularization",
   IEEE Transactions on Biomedical Engineering, vol. 52, no. 4, pp. 652-663, April 2005.

"""
import sys
from itertools import islice

import numpy as np
import pandas as pd
from scipy import interpolate

from locan.data.locdata import LocData
from locan.data.metadata_utils import _modify_meta
from locan.data.transform.transformation import transform_affine


__all__ = ['bunwarp']


def _unwarp(points, matrix_x, matrix_y, pixel_size):
    """
    Transform points with raw matrix from BunwarpJ.

    Parameters
    ----------
    points : array-like with shape (n_points, 2)
        Point coordinates to be transformed
    matrix_x, matrix_y : array-like
        Transformation matrix for x and y coordinates
    pixel_size : tuple (2,)
        Pixel size for x and y component as used in ImageJ for registration

    Returns
    -------
    numpy.ndarray with shape (n_points, 2)
        Transformed point coordinates
    """
    points_ = np.asarray(points)
    point_indices = np.divide(points_, pixel_size)

    matrix_size = matrix_x.shape
    assert matrix_size == matrix_y.shape

    x = np.arange(matrix_size[0])
    y = np.arange(matrix_size[1])
    z_x = matrix_x.T
    f_x = interpolate.interp2d(x, y, z_x, kind='linear')
    z_y = matrix_y.T
    f_y = interpolate.interp2d(x, y, z_y, kind='linear')

    new_points = np.array(
        [np.concatenate((f_x(*pind), f_y(*pind))) for pind in point_indices]
    )
    new_points = np.multiply(new_points, pixel_size)
    return new_points


def _read_matrix(path):
    """
    Read file with raw matrix from BunwarpJ.

    Parameters
    ----------
    path : str, bytes, os.PathLike
        Path to file with a raw matrix from BunwarpJ.

    Returns
    -------
    numpy.ndarray
        x transformation array
    numpy.ndarray
        y transformation array
    """
    with open(path) as file:
        header = list(islice(file, 2))

    # Get image heigth and width
    width = int(header[0].split("=")[1])
    height = int(header[1].split("=")[1])
    matrix_size = np.array([width, height])

    matrix_x = pd.read_csv(path, skiprows=4, header=None, nrows=matrix_size[1],
                                 delim_whitespace=True).values.T      # transform values to get array[x, y]
    matrix_y = pd.read_csv(path, skiprows=(6 + matrix_size[1]), header=None,
                                 delim_whitespace=True).values.T      # transform values to get array[x, y]

    return matrix_x, matrix_y


def bunwarp(locdata, matrix_path, pixel_size, flip=False):
    """
    Transform coordinates by applying a B-spline transformation
    as represented by a raw transformation matrix from BunwarpJ.

    Parameters
    ----------
    locdata : LocData
        specifying the localization data on which to perform the manipulation.
    matrix_path :
        Path to file with a raw matrix from BunwarpJ.
    pixel_size : tuple
        Pixel sizes used to determine transition matrix in ImageJ
    flip : bool
        Flip locdata along x-axis before transformation

    Returns
    -------
    locdata : LocData
        New localization data with transformed coordinates.
    """
    local_parameter = locals()

    matrix_x, matrix_y = _read_matrix(matrix_path)

    if flip:
        image_size = np.multiply(matrix_x.shape, pixel_size)
        locdata = transform_affine(locdata,
                                   matrix=[[-1, 0], [0, 1]],
                                   offset=[image_size[0], 0]
                                   )

    new_points = _unwarp(locdata.coordinates, matrix_x, matrix_y, pixel_size)

    # new LocData object
    new_dataframe = locdata.data.copy()
    df = pd.DataFrame({'position_x': new_points[:, 0], 'position_y': new_points[:, 1]},
                      index=locdata.data.index)
    new_dataframe.update(df)
    new_locdata = LocData.from_dataframe(new_dataframe)

    # update metadata
    meta_ = _modify_meta(locdata, new_locdata, function_name=sys._getframe().f_code.co_name,
                         parameter=local_parameter,
                         meta=None)
    new_locdata.meta = meta_

    return new_locdata



"""
Functions to compute locdata properties.
"""
from collections import namedtuple
import warnings
import logging

import numpy as np
from scipy.spatial.distance import pdist
from shapely.geometry import Point

from locan.data.region import Region, Region2D, RoiRegion


__all__ = ['distance_to_region', 'distance_to_region_boundary', 'max_distance', 'inertia_moments']

logger = logging.getLogger(__name__)


def distance_to_region(locdata, region):
    """
    Determine the distance to the nearest point within `region` for all localizations.
    Returns zero if localization is within the region.

    Parameters
    ----------
    locdata : LocData
        Localizations for which distances are determined.

    region : Region
        Region from which the closest point is selected.

    Returns
    --------
    numpy.ndarray
        Distance for each localization.
    """
    distances = np.full(len(locdata), 0.)
    if isinstance(region, (Region2D, RoiRegion)):
        for i, point in enumerate(locdata.coordinates):
            distances[i] = Point(point).distance(region.shapely_object)
    else:
        raise NotImplementedError("Region must be Region2D object.")

    return distances


def distance_to_region_boundary(locdata, region):
    """
    Determine the distance to the nearest region boundary for all localizations.
    Returns a positive value regardless of weather the point is within or outside the region.

    Parameters
    ----------
    locdata : LocData
        Localizations for which distances are determined.

    region : Region
        Region from which the closest point is selected.

    Returns
    --------
    numpy.ndarray
        Distance for each localization.
    """
    distances = np.full(len(locdata), 0.)
    if isinstance(region, (Region2D, RoiRegion)):
        for i, point in enumerate(locdata.coordinates):
            distances[i] = Point(point).distance(region.shapely_object.boundary)
    else:
        raise NotImplementedError("Region must be Region2D object.")

    return distances


def max_distance(locdata):
    """
    Return maximum of all distances between any two localizations in locdata.

    Parameters
    ----------
    locdata : LocData
        Localization data

    Returns
    -------
    dict
        A dict with key `max_distance` and the corresponding value being the maximum distance.
    """
    points = locdata.convex_hull.vertices
    D = pdist(points)
    distance = np.nanmax(D)
    return {'max_distance': distance}


def inertia_moments(points):
    """
    Return inertia moments (or principal components) and related properties for the given points.
    Inertia moments are represented by eigenvalues (and corresponding eigenvectors) of the covariance matrix.
    Variance_explained represents the eigenvalues normalized to the sum of all eigenvalues.
    For 2-dimensional data, orientation is the angle (in degrees) between the principal axis
    with the largest variance and the x-axis.
    Also for 2-dimensional data, eccentricity is computed as e=Sqrt(1-M_min/M_max).

    Parameters
    ----------
    points : tuple, list, numpy.ndarray of shape (n_points, n_dim)
        Point coordinates or other data.

    Returns
    -------
    namedtuple(InertiaMoments, 'eigenvalues eigenvectors variance_explained orientation eccentricity')
        A tuple with eigenvalues, eigenvectors, variance_explained, orientation, eccentricity

    Note
    ----
    The data is not standardized.
    """
    points = np.asarray(points)
    covariance_matrix = np.cov(points.T)
    eigen_values, eigen_vectors = np.linalg.eig(covariance_matrix)
    variance_explained = [eigen_value / sum(eigen_values) for eigen_value in eigen_values]
    index_max_eigen_value = np.argmax(eigen_values)

    if np.shape(points)[1] == 2:
        orientation = np.degrees(
            np.arctan2(eigen_vectors[1][index_max_eigen_value], eigen_vectors[0][index_max_eigen_value]))
        eccentricity = np.sqrt(1 - np.min(eigen_values) / np.max(eigen_values))
    else:  # todo implement for 3d
        logger.warning("Orientation and eccentricity have not yet been implemented for 3D.")
        orientation = np.nan
        eccentricity = np.nan

    InertiaMoments = namedtuple('InertiaMoments',
                                'eigenvalues eigenvectors variance_explained orientation eccentricity')
    inertia_moments = InertiaMoments(eigen_values, eigen_vectors, variance_explained, orientation, eccentricity)
    return inertia_moments



import pytest
import numpy as np
import pandas as pd
from locan import LocData
from locan.data.properties import statistics, ranges, range_from_collection


@pytest.fixture()
def locdata_simple():
    dict_ = {
        'position_x': [0, 0, 1, 4, 5],
        'position_y': [0, 1, 3, 4, 1]
    }
    return LocData(dataframe=pd.DataFrame.from_dict(dict_))


def test_statistics(locdata_simple):
    stat = statistics(locdata=locdata_simple)
    assert stat == {'position_x_count': 5.0, 'position_x_min': 0.0, 'position_x_max': 5.0, 'position_x_mean': 2.0,
                    'position_x_median': 1.0, 'position_x_std': 2.3452078799117149,
                    'position_x_sem': 1.0488088481701516, 'position_y_count': 5.0, 'position_y_min': 0.0,
                    'position_y_max': 4.0, 'position_y_mean': 1.8, 'position_y_median': 1.0,
                    'position_y_std': 1.6431676725154984, 'position_y_sem': 0.73484692283495345}


def test_statistics_from_dataframe(locdata_simple):
    stat = statistics(locdata=locdata_simple.data)
    assert stat == {'position_x_count': 5.0, 'position_x_min': 0.0, 'position_x_max': 5.0, 'position_x_mean': 2.0,
                    'position_x_median': 1.0, 'position_x_std': 2.3452078799117149,
                    'position_x_sem': 1.0488088481701516, 'position_y_count': 5.0, 'position_y_min': 0.0,
                    'position_y_max': 4.0, 'position_y_mean': 1.8, 'position_y_median': 1.0,
                    'position_y_std': 1.6431676725154984, 'position_y_sem': 0.73484692283495345}


def test_statistics_with_one_statfunction(locdata_simple):
    stat = statistics(locdata=locdata_simple, statistic_keys='min')
    assert stat == {'position_x_min': 0, 'position_y_min': 0}


def test_statistics_from_Series(locdata_simple):
    stat = statistics(locdata=locdata_simple.data['position_x'], statistic_keys='mean')
    assert stat['position_x_mean'] == 2
    stat = statistics(locdata=locdata_simple.data['position_x'], statistic_keys=('min', 'max'))
    assert stat == {'position_x_min': 0, 'position_x_max': 5}


def test_statistics_with_new_column(locdata_simple):
    locdata_simple.dataframe = locdata_simple.dataframe.assign(new=np.arange(5))
    stat = statistics(locdata=locdata_simple)
    # print(stat)
    assert stat == {'position_x_count': 5.0, 'position_x_min': 0.0, 'position_x_max': 5.0, 'position_x_mean': 2.0,
                    'position_x_median': 1.0, 'position_x_std': 2.3452078799117149,
                    'position_x_sem': 1.0488088481701516, 'position_y_count': 5.0, 'position_y_min': 0.0,
                    'position_y_max': 4.0, 'position_y_mean': 1.8, 'position_y_median': 1.0,
                    'position_y_std': 1.6431676725154984, 'position_y_sem': 0.73484692283495345,
                    'new_count': 5.0, 'new_min': 0.0, 'new_max': 4.0, 'new_mean': 2.0, 'new_median': 2.0,
                    'new_std': 1.5811388300841898, 'new_sem': 0.70710678118654757}


def test_range_from_collection(locdata_3d):
    collection = LocData.from_chunks(locdata_3d, n_chunks=2)
    result = range_from_collection(collection.references)
    assert result.position_x.min == 1
    assert result.position_x.max == 5
    assert result.position_z.min == 1
    assert result.position_z.max == 5


def test__ranges(locdata_2d):
    ranges_ = ranges(locdata_2d)
    assert np.array_equal(ranges_, [[1, 5], [1, 6]])
    ranges_ = ranges(locdata_2d, loc_properties=True)
    assert np.array_equal(ranges_, [[1, 5], [1, 6], [1, 6], [80, 150]])
    ranges_ = ranges(locdata_2d, loc_properties='intensity')
    assert np.array_equal(ranges_, [[80, 150]])
    ranges_ = ranges(locdata_2d, loc_properties=['intensity'])
    assert np.array_equal(ranges_, [[80, 150]])
    ranges_ = ranges(locdata_2d, loc_properties=['frame', 'intensity'])
    assert np.array_equal(ranges_, [[1, 6], [80, 150]])
    ranges_ = ranges(locdata_2d, loc_properties=('frame', 'intensity'))
    assert np.array_equal(ranges_, [[1, 6], [80, 150]])
    ranges_ = ranges(locdata_2d, loc_properties=None, special='zero')
    assert np.array_equal(ranges_, [[0, 5], [0, 6]])
    ranges_ = ranges(locdata_2d, loc_properties=True, special='link')
    assert np.array_equal(ranges_, [[1, 150],  [1, 150], [1, 150], [1, 150]])


def test__ranges_empty(locdata_empty, locdata_single_localization):
    assert ranges(locdata_empty) is None
    ranges_ = ranges(locdata_single_localization, loc_properties=None)
    assert np.array_equal(ranges_, [[1, 2], [1, 2]])
    ranges_ = ranges(locdata_single_localization, loc_properties=True)
    assert np.array_equal(ranges_, [[1, 2], [1, 2], [1, 2], [1, 2]])
